!******************************************************************************
!
! File : hamiltonian.f90
!   by : Alan Tackett
!   on : 02/07/97
!  for : PAW
!
! Module for calculating the Hamiltonian and Cohesive Energy of the PAW
! system.  Also contains routines to do H*Psi and O*Psi.
!
!******************************************************************************

Module hamiltonian

Use anderson_mixing
Use atom_data
Use basis_lib
Use coulomb_pack
Use crystal_data
Use crystal_symmetry
Use denvhat_pack
Use doijmatrix
Use exchange_corr
Use gausslib
Use gpoints 
Use laplacian
Use mathlib
Use mem_data
Use memmgr
Use misc
Use oinverse
Use options_data
Use orbital_pack
Use paw_inout
Use projectors
Use spherical_harmonic
Use psilib
Use search_sort
Use vhartree_pack
Use word
Use work_mgr

!**Use atom_charge

Implicit NONE!!!!

Integer, Pointer :: BandMap(:,:)

Real  :: Diag_Energy       !** K+V_ion energy
Real  :: hat_Energy        !** V^ + E^ energy
Real  :: Hartree_Energy   !** Hartree  energy
Real  :: OneExc_Energy     !** Exchange-correlation energy in AccumVxc
Real  :: TildeKE_Energy    !** KE energy for smooth hamiltonian
Real  :: TildePot_Energy   !** Pot energy for smooth hamiltonian
Real  :: TildeXC_Energy    !** XC energy for smooth hamiltonian
Real  :: Prev_CohesiveEnergy(2)

Complex, Pointer ::  Work_Dij(:)  !** Work array for mixing Dij's

!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
!  RestrictPot - Restricts the potential from the High planewave grid to the
!       low cutoff grid in FFT space.
!
!  Vin    - High PW grid in FFT space(Input)
!  Vout   - Low PW grid in FFT space(Returned) 
!
!******************************************************************************

Subroutine RestrictPot( Vin, Vout)
  Complex, Intent(IN) :: Vin(:)
  Complex, Intent(OUT) :: Vout(:)

  Integer :: i, N_low, N_high
  real    :: weight, Ec

  Ec = PW_Gcut(G_PROJ)*(1-V_Smooth_width)

  N_low = Gpnt_Size(G_PROJ) - 1
  N_High = Gpnt_Size(G_HIGH) - 1

  Vout = 0

Write(Log_Unit,*) 'RestrictPot: PW_Gcut=',PW_Gcut, ' * V_Smooth_Width=',V_Smooth_Width
  Vout(FFTmap_PROJ(1)) = Vin(FFTmap_High(1))
  Do i=2, Gpnt_size(G_PROJ)
!     If (V_Smooth_Width > 0.0) then
!        weight = abs(Gpnt(4,i)-PW_Gcut(G_PROJ))/(PW_Gcut(G_Proj)*V_Smooth_Width)
!        weight = 2.0/(exp(-weight) + 1.0) - 1
!     else
!        Weight = 1
!     End if

     Weight = 1
     If (Gpnt(4,i) > Ec) then
        Weight = Pi*(PW_Gcut(G_PROJ) - Gpnt(4,i)) / &
             (2*V_Smooth_Width*PW_Gcut(G_PROJ))
        Weight = sin(weight)**2
     End if
!Write(Log_Unit,*) 'RestrictPot: i=',i, ' * |G|=',Gpnt(4,i), ' * Wt=',weight 
!weight = 1
     Vout(FFTmap_HIGH(i)) = Vin(FFTmap_High(i)) * Weight
     Vout(FFTmap_HIGH(i+N_HIGH)) = Vin(FFTmap_High(i+N_High)) * Weight
  End Do

  Return
End Subroutine


!******************************************************************************
!
!  SymRho - Symmetrizes the density based on the crystal symmetry operations
!
!  Rho   - The Density to symmetrize
!
!  ** Uses plane waves **
!******************************************************************************

Subroutine SymRho( Den, Work)
  Complex , Intent(INOUT) :: Den(:)
  Complex , Intent(INOUT) :: Work(:)

  Integer :: i,j,k1,s
  Real  :: R(3), Rp(3), gcut, mag, arg
  Complex  :: zsum, zterm
  Integer :: Nx, Ny, Nz,  Gi(3), N(3), M1(3), M2(3)

  Call Start_Timer(Timer(SymRho_Timer))

  Nx = FFT_Grid(1,G_HIGH)
  Ny = FFT_Grid(2,G_HIGH)
  Nz = FFT_Grid(3,G_HIGH)

  !Work = 0
  !Call PerformFFT(FFT_TO_G, G_HIGH, Den, Work)
  !write(6,*) 'after PerformFFT', Work(1)

  Work = Den
  Call PerformFFT(FFT_TO_G, G_HIGH,  Work)
  write(Log_Unit,*) 'SymRho: Start Q=', Work(1)

  Den = 0

  Den(1) = Work(1)

  Do i=2, Gpnt_Size(G_HIGH)
        Gi = Npnt(:,i)

        Zsum = 0 
        Do j=1, Xtal%Rot_Size
           N = MatMul(Igrot(:,:,j),Gi)
           arg = DOT_PRODUCT(N, xtal%Trans(:,j))
           If (ABS(Arg) > 1E-8) then
              arg = Two_PI*arg
              zterm = CMPLX(cos(arg),sin(arg))
           else
              zterm = CMPLX(1,0)
           End If
           
           M1 = N + 1
           Where (M1<=0) M1 = M1 + FFT_Grid(1:3,G_High)
           IF ((M1(1)<=Nx) .AND. (M1(1)>0) .AND. &
                (M1(2)<=Ny) .AND. (M1(2)>0) .AND. &
                (M1(3)<=Nz) .AND. (M1(3)>0)) then
                 k1 = CalcLinearIndex(FFT_Grid(:,G_HIGH), M1)+1
                 zsum = zsum + Work(k1)*zterm
             !write(6,'("SymRho",12i4)') i,j,Gi,N,M1,k1
            else
              write(error_unit,*) 'Symrho: bad rotation',j,Gi,N
              STOP
           End If
        End Do

           zsum = zsum/xtal%Rot_Size
           Den(FFTmap_High(i)) = zsum
             !write(6,*) 'sum',FFTmap_High(i),zsum
           Den(FFTmap_High(i+Gpnt_Size(G_High)-1)) = CONJG(zsum)
             !write(6,*) 'sum*',FFTmap_High(i+Gpnt_Size(G_HIGH)-1),CONJG(zsum)
  End Do

  write(Log_Unit,*) 'SymRho: After SymRho Q=', Den(1)
  Call PerformFFT(FFT_TO_R, G_HIGH, Den)

!  write(6,*) 'in SymRho sum', sum(den)

  Call Stop_Timer(Timer(SymRho_Timer))

  Return
End Subroutine


!******************************************************************************
!
!  SymWij - Symmetrizes the Wij coefficients
!
!******************************************************************************

Subroutine SymWij
  type Wij_Atom
     Complex, Pointer :: Wij(:,:)
  End Type

  Integer :: i,j,k, nili, njlj, Li, Lj,mi1, mj1, mi2, mj2, Basis_Size
  Integer :: ibase, jbase
  Real    :: c
  Complex :: zsum
  Complex, Pointer :: Wij(:,:), Wij_Copy(:,:)
  Type (Wij_Atom) :: Wij_List(Specific_Atoms)
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  Character*100 :: Msg


  c = 1.0/XTal%Rot_Size

  Do i=1, Specific_Atoms          !**** Make a temp copy of Wij ****
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     
     j=AT%nlm_Size
     Allocate(Wij_List(i)%Wij(j, j), STAT=k)
     Write(MSg,*) "SymWij: Error Allocating Temp array for atom:",i
     Call Check_Error(k, Msg, Error_Unit, .TRUE., IDTEXT="SymWij:")
     Wij_List(i)%Wij = A%Wij
  End Do

  Do i=1, Specific_Atoms        !*** Symmetrize Wij
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     
     Basis_Size = AT%Basis_Size     
     Wij => A%Wij

     Do nili=1, Basis_Size
        Li = AT%L_Value(nili)
        ibase = AT%nl_Base(nili)+Li
        Do mi1=-Li, Li
           Do njlj=1, Basis_Size
              Lj = AT%L_Value(njlj)
              jbase = AT%nl_Base(njlj)+Lj
              Do mj1=-Lj, Lj

                 zsum = 0
                 Do j=1, XTal%Rot_Size
                    Wij_Copy => Wij_List(iatrans(i,j))%Wij

                    Do mi2=-Li, Li
                       Do mj2=-Lj, Lj
                         zsum= zsum + (zarot(mi2+Li+1,mi1+Li+1,Li+1,j)) &
                               *CONJG(zarot(mj2+Lj+1,mj1+Lj+1,Lj+1,j)) &
                               *Wij_copy(ibase+mi2, jbase+mj2)
                       End Do
                    End Do

                 End Do 
                 Wij(ibase+mi1, jbase+mj1) = zsum*c
              End Do
           End Do    !**njlj
        End Do
     End Do    !**nili
  End Do


  Do i=1, Specific_Atoms        !*** Free Wij Copy
     DeAllocate(Wij_List(i)%Wij)
  End Do

  Return
End Subroutine


!******************************************************************************
!
!  CalcWij - Calculates the Wij coefficients
!
!  PDot - Projector times Psi products. The indices of PDot are as follows:
!
!       PDOT(atom, +/-M, ProjIndex, Band, Kpnt)
!          
!       Where 1=+M and 2=-M
!      
!
!******************************************************************************

Subroutine CalcWij(PDot)
  Complex, Intent(IN) :: PDot(:,:,:)
  complex :: sum1,sum2 ! test
  Integer :: Base, i, j, a
  Complex, Pointer :: Wij(:,:)

  sum1=0  !test
  Do a=1, Specific_Atoms
     Wij => Atom_List(a)%Wij

     Wij = 0

     Base = PLM_AtomRange(1,a) - 1
     Do i=PLM_AtomRange(1,a), PLM_AtomRange(2,a)
        Do j=PLM_AtomRange(1,a), PLM_AtomRange(2,a)
           Wij(i-Base, j-Base) = SUM(CONJG(PDot(i,:,:))*PDot(j,:,:)*Occupancy)
        End Do
     End Do

  i=AtomType_Info(Atom_List(a)%TypeIndex)%nlm_Size   !!testing
!  sum2=SUM(Wij(1:i,1:i)*AtomType_Info(Atom_List(a)%TypeIndex)%Oij(1:i,1:i))
 
!  write(Log_Unit,*) 'test Wij for atom',a,sum2   !!testing
!  sum1=sum1+sum2   !!testing


  End Do
!  write(Log_Unit,*) 'total wij*Oij test', sum1 !! testing
  
  Return
End Subroutine

!******************************************************************************
!                            a
!  CalcQlm - Calculates the Q  coefficients according to eq. A22 and 
!                            LM 
!
!             also accumulates the Qlm term in eq. A28 to dEdQlm
!
!******************************************************************************

Subroutine CalcQlm(energy)
  Real, Intent(INOUT) :: Energy
  
  Integer        :: i, j, k, Qlm_Size, Offset_Q
  Integer        :: nili, njlj, mi, mj, M, L, Li, Lj, ibase, jbase
  Integer, Pointer :: LUT_Orb(:), nl_Base(:), nlm_LUT(:,:)
  Real, Pointer  :: aQlm(:), Hat(:)
  Complex, Pointer :: Qlm(:,:), Wij(:,:), dEdQlm(:,:)
  Real     :: fn,tt
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  
     tt=0    !!!!!!
  Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)

     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     aQlm => AT%aQlm
     LUT_Orb => AT%LUT_Orb
     Qlm_Size = AT%QVlm_Size
     Qlm => A%Qlm
     dEdQlm => A%dEdQlm
     Wij => A%Wij
     Offset_Q = 2*MaxVal(AT%L_Value) + 1
     !Qlm(1, Offset_Q) = Qlm(1,Offset_Q) + AT%Core_Charge - At%Atomic_Charge 
     Qlm(1, Offset_Q) = Qlm(1,Offset_Q) + AT%Qeffion
     Hat => AT%Hat_SelfEnergy

     Do j=1, Qlm_Size
        Call Orbital_Decode(LUT_Orb(j), nili, njlj, mi, mj, L)
        M = mj - mi
        ibase = nl_base(nili);  ibase = ibase + nlm_LUT(2, ibase)
        jbase = nl_base(njlj);  jbase = jbase + nlm_LUT(2, jbase)

        Qlm(L+1, M+Offset_Q) = Qlm(L+1, M+Offset_Q) + &
                 Wij(ibase+mi, jbase+mj) * aQlm(j)
     End Do

     !***** Now accum dEdQlm and the Energy *****
     Do L=0, Offset_Q - 1

        fn = Hat(L)

        Do M=-L, L
           dEdQlm(L+1, Offset_Q+M) = dEdQlm(L+1, Offset_Q+M) - &
                     2*fn * CONJG(Qlm(L+1,Offset_Q + M))

           Energy= Energy - &
                fn*CONJG(Qlm(L+1,Offset_Q+M))*Qlm(L+1,Offset_Q+M)
        End Do
     End Do
   tt=tt+Qlm(1,Offset_Q) !!!!!!
  ! add coretail effects
  Energy=Energy-AT%CoreTail_SelfEnergy-Qlm(1,Offset_Q)*AT%CoreTail_HatEnergy
  dEdQlm(1,Offset_Q)=dEdQlm(1,Offset_Q)-AT%CoreTail_HatEnergy
  End Do

  write(Log_Unit,*) ' check Q00 ', tt   !!!!!

!   Write(6,*) ' dEdQlm with self term'
!   call WriteQlm(2,6)


  Return
End Subroutine

!******************************************************************************
!                  
!  AccumVhartree - Accumulates the Vhartree terms in Dij based on eq A26
!                  
!******************************************************************************

Subroutine AccumVhartree(Energy)
  Real, Intent(INOUT) :: Energy
  
  Integer        :: i, j, k, Qlm_Size, Offset_Q,  Cijkl_Size
  Integer        :: nili, njlj, nlll,nklk, mi, mj, mk, ml
  Integer        :: ibase,jbase,lbase,kbase
  Integer, Pointer :: LUT_Cijkl(:), nl_base(:), nlm_LUT(:,:)
  Real, Pointer  :: Cijkl(:)
  Complex, Pointer :: Dij(:,:), Wij(:,:)
  Complex        :: fn
  Real           :: Venergy
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  
  Venergy = 0

  Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)

     Offset_Q = 2*MaxVal(AT%L_Value) + 1

     Cijkl => AT%Cijkl
     LUT_Cijkl => AT%LUT_Cijkl
     Cijkl_Size = AT%Cijkl_Size
     nl_base => AT%nl_base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij

     Do j=1, Cijkl_Size
        Call Coulomb_Decode(LUT_Cijkl(j), nili, njlj, nklk, nlll, mi, mj, mk)
        ml = mi - mj + mk
        ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase) + mi;
        jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase) + mj;
        kbase = nl_Base(nklk); kbase = kbase + nlm_lut(2,kbase) + mk;
        lbase = nl_Base(nlll); lbase = lbase + nlm_lut(2,lbase) + ml;

        fn = Wij(kbase, lbase) * Cijkl(j)
        Dij(ibase,jbase) = Dij(ibase,jbase) + fn             

        Venergy = Venergy + Wij(ibase,jbase) * fn
     End Do
  End Do

  Energy = Energy + 0.5*Venergy

  Return
End Subroutine

!******************************************************************************
!                  
!  AccumVhatdEdQlm - Accumulates the Vhat terms in Dij based on eq A23 
!                    and also accumulates the contributions to dEdQlm
!                  
!******************************************************************************

Subroutine AccumVhatdEdQlm(Energy)
  Real, Intent(INOUT) :: Energy
  
  Integer        :: i, j, k, Vlm_Size, Offset_Q, ibase,jbase
  Integer        :: nili, njlj, mi, mj, M, L
  Integer, Pointer :: LUT_Orb(:), nl_base(:), nlm_LUT(:,:)
  Real, Pointer  :: aVlm(:)
  Complex, Pointer :: dEdQlm(:,:), Qlm(:,:), Wij(:,:), Dij(:,:)
  Complex          :: fn
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT


  Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)

     Offset_Q = AT%Hat_MaxL + 1

     aVlm => AT%aVlm
     LUT_Orb => AT%LUT_Orb
     Vlm_Size = AT%QVlm_Size
     Qlm => A%Qlm
     dEdQlm => A%dEdQlm
     nl_base => AT%nl_base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij
     
     Do j=1, Vlm_Size
        Call Orbital_Decode(LUT_Orb(j), nili, njlj, mi, mj, L)
        M = -(mj - mi)           !** Change sign for aVlm
        
        ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase) + mi;
        jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase) + mj;

        fn = aVlm(j) * Qlm(L+1, M+Offset_Q)   !** Reverse M-sign for Qlm
        Dij(ibase, jbase) = Dij(ibase,jbase) - fn

        Energy = Energy - Real(Wij(ibase,jbase) * fn)

                    !** Reverse M-sign for dEdQlm
        dEdQlm(L+1, Offset_Q + M) = dEdQlm(L+1, Offset_Q + M) - &
             Wij(ibase,jbase) * aVlm(j)
     End Do
  End Do

!   Write(6,*) ' dEdQlm with vhat term'
!   call WriteQlm(2,6)

  Return
End Subroutine


!******************************************************************************
!                  
!  CalcVlocal - Calculates Vlocal                        
!                  
!******************************************************************************

Subroutine CalcVlocal(V)
  Complex,         Intent(INOUT) :: V(:)

  Real    :: theta, Gmag, RadG, G(3), Esum
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi
  Type (Atom_Info_Fixed), Pointer :: AT

  V = 0

  RadIndex = 0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     Gmag = Gpnt(4,gindex)

     !*** If needed inc the RadIndex ***
     If (G_NewMag(gindex)) RadIndex = RadIndex + 1

     Vsum = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)
        RadG = AT%Rad_VLocal(RadIndex)

        !*** Create the structure factors and accumulate vlocal***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
        
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase = CMPLX(cos(theta), -sin(theta))
!Phase = CONJG(Phase)
           Vsum = Vsum + Phase*RadG
        End Do
     End do

     V(gindex) = V(gindex) + Vsum
  End Do

  Return
End Subroutine



!******************************************************************************
!                  
!  AccumVlocal - Accumulates the Vlocal                        
!                  
!******************************************************************************

Subroutine AccumVlocal(Energy, Den, V)
  Complex,         Intent(INOUT) :: V(:)
  Complex,         Intent(IN) :: Den(:)
  Real,            Intent(INOUT) :: Energy !Accumulate Pot. Energy

  Real    :: theta, Gmag, RadG, G(3), Esum
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi
  Type (Atom_Info_Fixed), Pointer :: AT


  Esum = 0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     Gmag = Gpnt(4,gindex)

     fi = FFTmap_High(gindex)
     Vsum = SCFvalues%V_local(gindex)
     V(fi) = V(fi) + Vsum

     Esum = Esum + Vsum*Conjg(Den(fi))

     If (gindex>1) then
        fi = FFTmap_High(Gindex + Gpnt_Size(G_HIGH)-1)
        V(fi) = V(fi) + CONJG(Vsum)
        Esum = Esum + Conjg(Vsum*Den(fi))
     End if
  End Do

  Energy = Energy + Esum

  Return
End Subroutine


!******************************************************************************
!                  
!  AccumVlocal - Accumulates the Vlocal                        
!
!***OLD**                  
!******************************************************************************

Subroutine AccumVlocal_OLD(Energy, Den, V)
  Complex,         Intent(INOUT) :: V(:)
  Complex,         Intent(IN) :: Den(:)
  Real,            Intent(INOUT) :: Energy !Accumulate Pot. Energy

  Real    :: theta, Gmag, RadG, G(3), Esum
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi
  Type (Atom_Info_Fixed), Pointer :: AT

  Esum = 0
  RadIndex = 0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     Gmag = Gpnt(4,gindex)

     !*** If needed inc the RadIndex ***
     If (G_NewMag(gindex)) RadIndex = RadIndex + 1

     Vsum = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)
        RadG = AT%Rad_VLocal(RadIndex)

        !*** Create the structure factors and accumulate vlocal***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
        
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase = CMPLX(cos(theta), -sin(theta))
           Vsum = Vsum + Phase*RadG
        End Do
     End do

     fi = FFTmap_High(gindex)
     V(fi) = V(fi) + Vsum

     Esum = Esum + Vsum*Conjg(Den(fi))

     if (gindex.le.30) then
       write(6,'("vloc",2i6,1p6e10.3)') gindex,fi,Gmag,Vsum,Conjg(Den(Fi)),Esum
     endif


     If (gindex>1) then
        fi = FFTmap_High(Gindex + Gpnt_Size(G_HIGH)-1)
        V(fi) = V(fi) + CONJG(Vsum)
        Esum = Esum + Conjg(Vsum*Den(fi))

     if (gindex.le.30) then
       write(6,'("vloc",2i6,1p6e10.3)') gindex,fi,Gmag,Vsum,Conjg(Den(Fi)),Esum
     endif
     End if


  End Do

   Energy = Energy + Esum

   Write(6,*) ' In AccumVlocal ', Esum, Energy

  Return
End Subroutine


!******************************************************************************
!                  
!  CalcDenhat - Accumulates the n^ density
!                  
!******************************************************************************

Subroutine CalcDenHat( DenHat)
  Complex, Intent(INOUT) :: DenHat(:)    ! input density is augmented with n^
  
  Real    :: theta, Gmag,  RadG, MinQlm, G(3)
  Complex :: Phase(Specific_Atoms), Ylm(13), Ylm_neg(13), den,tmp
  Complex, parameter :: ai=(0,-1)
  Integer :: atom, atype, L, m, lut, gindex, Offset_Q, HatIndex, fi, Gi(3)
  Logical :: WhichL(20, Atom_Types)
  Type (Atom_Info_Fixed), Pointer :: AT
  Type (Specific_Atom), Pointer :: A

  Call Start_Timer(Timer(DenHat_Timer))
  !** First create L filter using the Qlm **
  WhichL = .FALSE.
  !MinQlm = 1.E-30
  Do atom=1, Specific_Atoms
     A => Atom_List(atom)
     AT => AtomType_Info(A%TypeIndex)

     MinQlm = 1E-20*MAXVAL(REAL(CONJG(A%Qlm)*A%Qlm))

     Offset_Q = AT%Hat_MaxL + 1
     Do L=0, AT%Hat_MaxL
        Do m=-L, L
           theta = CONJG(A%Qlm(L+1, Offset_Q+m))*A%Qlm(L+1, Offset_Q+m)
           If (theta > MinQlm) WhichL(L+1, A%TypeIndex) = .TRUE.
        End Do
     End Do
  End Do

  !*** Now calculate the Hat density in G-space ***
  HatIndex = 0
  Do gindex=1, Gpnt_Size(G_HIGH)
     fi = FFTmap_High(gindex)
     G = Gpnt(1:3,gindex)
     Gmag = Gpnt(4,gindex)

     !*** If needed inc the HatIndex ***
     If (G_NewMag(gindex)) HatIndex = HatIndex + 1

     den = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)

        Offset_Q = AT%Hat_MaxL + 1

        !*** Create and store the structure factors ***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
        
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase(atom) = CMPLX(cos(theta), -sin(theta))
        End Do


        Do L=0, AT%Hat_MaxL
           If (WhichL(L+1, atype)) then
              Ylm = Spharm(G(1), G(2), G(3), L, .TRUE.)
              
              RadG = AT%RadHat(HatIndex, L+1)
              Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
                 atom = Atomtype_map(lut)

!If (gindex==1) then 
!  Write(Log_Unit,*) 'CalcDenHat: atype=',atype, ' * L=', L, ' * RadG=',RadG, ' * Phase=',Phase(atom)
!end If
                 tmp=0
                 Do m=-L, L
                    tmp=tmp+Atom_List(Atom)%Qlm(L+1, Offset_Q+m)*Ylm(L+m+1)
!If (gindex==1) then 
!  Write(Log_Unit,*) 'CalcDenHat: atype=',atype, ' * atom=',atom, ' * L=',L, ' * m=',m, ' * Qlm=',Atom_List(Atom)%Qlm(L+1, Offset_Q+m), ' * Term=',Atom_List(Atom)%Qlm(L+1, Offset_Q+m) * Phase(atom)*RadG*Ylm(L+m+1)
!end If
                 Enddo

                 tmp=tmp*(ai**L)
                 den = den + tmp*Phase(atom)*RadG
!If (gindex==1) then
!   Write(Log_Unit,*) 'CalcDenHat: atom=',atom, ' * L=',L, ' * SumDenHat=',den
!End If
             End Do
           End if
        End Do
     End Do

!If (gindex == 1) Write(Log_Unit,*) 'CalcDenHat: Qhat=',Den, ' * Den(1)=',DenHat(1)

     DenHat(fi) = DenHat(fi) + den

! Store RhoHat
     SCFvalues%RhoHat(gindex)=den

     !write(6,*) 'CalcDen',fi,gmag,den
     If (gindex>1) then
        fi = FFTmap_High(gindex+Gpnt_size(G_HIGH)-1)
        DenHat(fi) = DenHat(fi) + CONJG(den)
     End If
  End Do

  Call Stop_Timer(Timer(DenHat_Timer))
    !write(6,*) 'in denhat -', denHat(1),denHat(7),denHat(97),denHat(200)
  Return
End Subroutine

!******************************************************************************
!                  
!  CalcDen_dEdQlm - Accumulates the n~+n^ density term to dE/dQlm.
!                  
!******************************************************************************

Subroutine CalcDen_dEdQlm( Den)
  Complex, Intent(IN) :: Den(:)     !This must be n~ + n^(G)
  
  Real    :: theta, Gmag,  RadG, MinQlm, c1, G(3)
  Complex :: Phase(Specific_Atoms), Ylm(13), Ylm_neg(13), DenG, tmp
  Complex , parameter :: ai=(0,-1)
  Integer :: atom, atype, L, m, lut, gindex, Offset_Q, HatIndex,Gi(3)
  Logical :: WhichL(20, Atom_Types)
  Type (Atom_Info_Fixed), Pointer :: AT
  Type (Specific_Atom), Pointer :: A


  c1 = 2*Four_Pi/xtal%Volume


  HatIndex = 1
  Do gindex=2, Gpnt_Size(G_HIGH)
     G = Gpnt(1:3,gindex)
     Gmag = Gpnt(4,gindex)

     DenG = CONJG(Den(FFTmap_High(gindex))+SCFValues%CoreTail(gindex))/(Gmag*Gmag)

     !*** If needed inc the HatIndex ***
     If (G_NewMag(gindex)) HatIndex = HatIndex + 1

     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)

        Offset_Q = AT%Hat_MaxL + 1

        !*** Create and store the structure factors ***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
        
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase(atom) = CMPLX(cos(theta), -sin(theta))
        End Do

        Do L=0, AT%Hat_MaxL
           Ylm = Spharm(G(1), G(2), G(3), L, .TRUE.)
           !Ylm_neg = Spharm(-G(1), -G(2), -G(3), L, .TRUE.)
           Ylm_neg = Ylm; if (L>0) Ylm_neg=Ylm_neg*((-1)**L)
           
           RadG = AT%RadHat(HatIndex, L+1)
           tmp = ai**L
           Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
              atom = Atomtype_map(lut)

              Do m=-L, L
                 Atom_List(atom)%dEdQlm(L+1, Offset_Q+m) = &
                      Atom_List(atom)%dEdQlm(L+1, Offset_Q+m) + &
                      c1 * tmp * (DenG * Ylm(L+m+1)*Phase(atom) + &
                      Ylm_neg(L+m+1)*CONJG(DenG * Phase(atom)))*RadG
              End Do
           End Do
        End Do
     End Do
  End Do

!   Write(6,*) ' dEdQlm with den term'
!   call WriteQlm(2,6)

  Return
End Subroutine


!******************************************************************************
!                  
!  CalcDenhat_dEdQlm - Accumulates the  n^ density in Den to n~ and 
!                also accumulates the term in dEdQlm containing n^+n~.
!                  
!******************************************************************************

Subroutine CalcDenHat_dEdQlm( Den)
  Complex,         Intent(INOUT) :: Den(:)


  Call CalcDenHat( Den)
  Call CalcDen_dEdQlm( Den)

  Return
End Subroutine

!******************************************************************************
!
!  AccumPotential - Accumulates the potential energy due to n~+n^+ncore~.
!
!******************************************************************************

Subroutine AccumPotential( Energy, Rho, V)
  Real,    Intent(INOUT) :: Energy
  Complex, Intent(INOUT) :: Rho(:)
  Complex, Intent(INOUT) :: V(:)

  Real :: c1, c2 , Esum
  Complex :: Vpnt, Den
  Integer :: gi, N

  c1 = Four_Pi/xtal%Volume
  c2 = 2*c1

 !write(6,*) 'in AccumPotential',c1,c2

  N = Gpnt_Size(G_HIGH) - 1

  Esum = 0
  Do gi=2, Gpnt_Size(G_High)
     Den = Rho(FFTmap_High(gi)) +SCFValues%CoreTail(Gi)
     Vpnt = c2*(Den)/ Gpnt(4,gi)**2
     Esum = Esum + REAL(Vpnt*CONJG(den))
     !write(6,*) 'ACCPOT',gi,ESUM
    !V(FFTmap_High(gi)) = V(FFTmap_High(gi)) + FFT_Grid(4,G_High)*Vpnt
    !V(FFTmap_High(gi+N)) = V(FFTmap_High(gi+N)) + FFT_Grid(4,G_High)*CONJG(Vpnt)
     If (V_MixType /= MIX_V) then
        V(FFTmap_High(gi)) = V(FFTmap_High(gi)) + Vpnt      !** Orig
        V(FFTmap_High(gi+N)) = V(FFTmap_High(gi+N)) + CONJG(Vpnt)
     else
        Rho(FFTmap_High(gi)) = Vpnt
        Rho(FFTmap_High(gi+N)) = CONJG(Vpnt)
      End if
  End Do
  
  Energy = Energy + Esum

  Return
End Subroutine

!******************************************************************************
!
!  AccumShift - Accumulates the coulomb shift term eq A27
!
!******************************************************************************

Subroutine AccumShift
  
  Integer        :: i, j, k, Qlm_Size, Offset_Q, ibase,jbase
  Integer        :: nili, njlj, mi, mj, M, L
  Integer, Pointer :: LUT_Orb(:), nlm_LUT(:,:), nl_Base(:)
  Real, Pointer  :: aQlm(:)
  Complex, Pointer :: dEdQlm(:,:), Dij(:,:)
  Type (Specific_Atom), Pointer :: A
  Type (Atom_Info_Fixed), Pointer :: AT
  
  Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)

     Offset_Q = AT%Hat_MaxL + 1

     aQlm => AT%aQlm
     LUT_Orb => AT%LUT_Orb
     Qlm_Size = AT%QVlm_Size
     dEdQlm => A%dEdQlm
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Dij => A%Dij

     Do j=1, Qlm_Size
        Call Orbital_Decode(LUT_Orb(j), nili, njlj, mi, mj, L)
        M = mj - mi

        ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase) + mi;
        jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase) + mj;

        Dij(ibase,jbase) = Dij(ibase,jbase) + aQlm(j) * dEdQlm(L+1, M+Offset_Q)
     End Do
  End Do

  Return
End Subroutine

!******************************************************************************
!
!  AddDiagElements - Adds the Diagonal Elements to the Dij for each atom
!                    eq A10, A11
!
!******************************************************************************

Subroutine AddDiagElements(Energy)
   Real, Intent(INOUT) :: Energy

   Integer :: i, j, k, t, Basis_Size, Overlap_Size
   Integer :: nili, njlj, m, mmin, mmax, Li, Lj, ibase,jbase
   Complex, Pointer :: Dij(:,:), Wij(:,:)
   Real, Pointer :: Vion(:), KE(:)
   Integer, Pointer :: L_Values(:), nlm_LUT(:,:),nl_base(:)
   Complex       :: fn
   Integer, Pointer :: LUT(:)

   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT
 
   Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     Dij => A%Dij
     Wij => A%Wij
     Vion => AT%V_ion
     KE => AT%Kinetic
     Basis_Size = AT%Basis_Size
     L_Values => AT%L_Value
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT

     t = 0
     Do nili=1, Basis_Size
        Li = L_Values(nili)
        ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase);

        Do njlj=1, nili
           Lj = L_Values(njlj)

           If (Li == Lj) then
              t = t + 1
              jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase);

              fn = KE(t) + Vion(t)

              Do m=-Li, Li
                 Dij(ibase+m,jbase+m) = Dij(ibase+m,jbase+m) + Fn
                 Energy = Energy+Wij(ibase+m,jbase+m)*Fn

                 if (nili /= njlj) then
                    Dij(jbase+m,ibase+m) = Dij(jbase+m,ibase+m) + Fn
                    Energy = Energy+Wij(jbase+m,ibase+m)*Fn   
                 End If
              End Do
           End If
        End Do
     End Do
  End Do

  Return
End Subroutine

!******************************************************************************
!                                                          ~a   a
!  AccumVxc_LDA - Accumulates the Exchange correlation pot for N + N  to Dij
!    for LDA in Perdew_Wang form only
!
!   R=0 is included!!!!!!!!!!
!
!******************************************************************************

Subroutine AccumVxc_LDA( Energy)
   Real,    Intent(INOUT) :: Energy

   Integer :: i,j,k, nili, njlj, mi, mj, Li, Lj, Rk, RS, count
   Integer ::nklk, nlll, mk, ml, Lk, Ll, alpha, Basis_Size, Rad_Size
   Integer :: ibase,jbase,kbase,lbase, Skip_Size, Rad_Index, Mesh_Size
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:),  Ylm(:,:,:)
   Real,    Pointer :: R2(:), Phi_ij_R2(:), TPhi_ij_R2(:)
   Real,    Pointer :: CoreXC(:), CoreDen(:),FnRad(:,:)
   Complex          :: Ylm_l(13), Ylm_k(13), ylm1, ylm2
   Complex, Pointer :: Ylm_i(:), Ylm_j(:)
   Real             ::  Vxc1, Exc1, Vxc2, Exc2, fnint, Esum, PhiRk, TPhiRk
   Complex          :: den_phi, den_Tphi, den
   Real             :: Efn,fn,Fnsum
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT
   Real             :: DR
   Real, Pointer    :: Cij(:,:)
   Character*100    :: Msg


   Call Start_Timer(Timer(AtomVxc_Timer))
!   Write(msg,*) 'AccumVxc: Error allocating Cij. Size=',Basis_Size
!   Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "AccumVxc:")
   
   Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij
     CoreDen => AT%Core_Density
     CoreXC => AT%Core_Xchange
     R2 => AT%R2
     Ylm => AT%Ylm
     Cij => AT%Cij
     FnRad => AT%FnRad
     Mesh_size = AT%Rad_Size


      DR = AT%Mesh_Step*AT%Rad_Skip

     Basis_Size = AT%Basis_Size

     FnSum = 0; ESum = 0
     Do alpha = 1, Num_AngPoints
        Cij = 0
        Do nili=1, Basis_Size     !*** Calulate Cij for n^a and n~^a
           Li = AT%L_Value(nili)
           Ylm_i => Ylm(:, Li+1, alpha)
           ibase = nl_Base(nili);  ibase = ibase + nlm_lut(2,ibase);      
           Do njlj=1, Basis_Size
              Lj = AT%L_Value(njlj)
              Ylm_j => Ylm(:, Lj+1, alpha)
              jbase = nl_Base(njlj);  jbase = jbase + nlm_lut(2,jbase);      
           
              den = 0
              Do mi = -Li, Li
                 ylm1 = CONJG(Ylm_i(mi+Li+1))
              
                 Do mj=-Lj, Lj
                    ylm2 = Ylm_j(mj+Lj+1)
                    den = den + Wij(ibase+mi,jbase+mj) * Ylm1 * Ylm2
                 End Do
              End Do

              Cij(nili, njlj) = Den
           End do
        End Do

        !**** Calculate the Radial Densities and store them *****
        FnRad = 0

        Count = 0
        Do nili=1, Basis_Size
           Do njlj=1, Basis_Size
              count = count + 1
              Phi_ij_R2 => AT%Phi_ij_R2(:,count)
              TPhi_ij_R2 => AT%TPhi_ij_R2(:,count)

              FnRad(1:Mesh_Size,1)=FnRad(1:Mesh_Size,1) &
                   + Cij(nili,njlj)*Phi_ij_R2(1:Mesh_Size)

              FnRad(1:Mesh_Size,2)=FnRad(1:Mesh_Size,2) &
                   + Cij(nili,njlj)*TPhi_ij_R2(1:Mesh_Size)
           End Do
        End Do

        !*** Add CoreDen  ***
        FnRad(1:Mesh_Size,1) = FnRad(1:Mesh_Size,1) + &
             CoreDen(1:Mesh_Size)
        FnRad(1:Mesh_Size,2) = FnRad(1:Mesh_Size,2) + &
             AT%CoreTail_Density(1:Mesh_Size)

             !**** Calculate Vxc and Exc for Ra*****
        Do Rk=1, Mesh_Size
          
           Call pwldafunc((FnRad(Rk,1)),Exc1,Vxc1)
           Call pwldafunc((FnRad(Rk,2)),Exc2,Vxc2)
           
           FnRad(Rk, 3:6) = (/Vxc1, Vxc2, Exc1, Exc2/)
        ! write(6,'(i5,1p7e15.7)') Rk,Vxc1,Vxc2,Exc1,Exc2
        End Do

        FnRad(1:Mesh_Size,7) =   R2(1:Mesh_Size) * &
             (FnRad(1:Mesh_Size,5)*FnRad(1:Mesh_Size,1) &
             - FnRad(1:Mesh_Size,6)*FnRad(1:Mesh_Size,2))

        !FnRad(1:Mesh_Size,7) = FnRad(1:Mesh_Size,7) - &
        !   R2(1:Mesh_Size) *  &
        !     CoreXC(1:Mesh_Size) *  CoreDen(1:Mesh_Size)

        Efn= IntSimpson(Mesh_Size, DR, FnRad(1:Mesh_Size,7)) * &
              AngPoints(4,alpha)*4*Pi
        Energy = Energy + Efn 
       !write(6,*) 'alpha efn energy',alpha,efn,energy

        Count = 0
        Do nili=1, Basis_Size
           Li = AT%L_Value(nili)
           Ylm_i => Ylm(:, Li+1, Alpha)
           ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase);


           Do njlj=1, Basis_Size
              Lj = AT%L_Value(njlj)
              Ylm_j => Ylm(:, Lj+1, Alpha)
              jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase);      
              count = count + 1

              Phi_ij_R2 => AT%Phi_ij_R2(:,count)
              TPhi_ij_R2 => AT%TPhi_ij_R2(:,count)

              !**** Radial Integral Portion ******
              FnRad(1:Mesh_Size, 7) =  &
                 (FnRad(1:Mesh_Size,3) * Phi_ij_R2(1:Mesh_Size) - &
                    FnRad(1:Mesh_Size,4) * TPhi_ij_R2(1:Mesh_Size)) * &
                     R2(1:Mesh_Size) 

              FnRad(1,7) = 0
              FnRad(Mesh_Size,7) = 0
              
              fn = IntSimpson(Mesh_Size, DR, FnRad(1:Mesh_Size,7)) * &
                     AngPoints(4,alpha)*Four_Pi

              Do mi=-Li, Li
                 Do mj=-Lj, Lj
                    Dij(ibase+mi,jbase+mj) = Dij(ibase+mi,jbase+mj) + &
                         fn * CONJG(Ylm_i(Li+Mi+1))*Ylm_j(Lj+mj+1)
                 End Do
              End Do  
           end Do !njlj
        End do !nili
     End Do

  End Do

  Call Stop_Timer(Timer(AtomVxc_Timer))

  Return
End Subroutine

!******************************************************************************
!                                                          ~a   a
!  AccumVxc_GGA - Accumulates the Exchange correlation pot for N + N  to Dij
!    for GGA in Perdew-Burke-Ezerholf form only
!
!******************************************************************************

Subroutine AccumVxc_GGA( Energy)
   Real,    Intent(INOUT) :: Energy

   Integer :: i,j,k, nili, njlj, mi, mj, Li, Lj, Rk, RS, count
   Integer ::nklk, nlll, mk, ml, Lk, Ll, alpha, Basis_Size, Rad_Size
   Integer :: ibase,jbase,kbase,lbase, Skip_Size, Rad_Index, Mesh_Size
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:),  Ylm(:,:,:)
   Complex, Pointer :: dYlmdtheta(:,:,:), dYlmdphi(:,:,:)
   Real,    Pointer ::  R2(:), Phi_ij_R2(:), TPhi_ij_R2(:)
   Real,    Pointer :: dPhijdr(:),dTPhijdr(:)
   Real,    Pointer :: CoreXC(:), CoreDen(:),FnRad(:,:),CoreGrad(:)
   Complex          :: ylm1, ylm2,ylm1t,ylm1p,ylm2t,ylm2p
   Complex, Pointer :: Ylm_i(:), Ylm_j(:)
   Complex, Pointer :: dYlmdt_i(:), dYlmdt_j(:)
   Complex, Pointer :: dYlmdp_i(:), dYlmdp_j(:)
   Real             ::  Vxc1, Exc1, Vxc2, Exc2, fnint, Esum, PhiRk, TPhiRk
   Real             ::  fxc,gxc,dfxcdn,dfxcdg
   Complex          :: den_phi, den_Tphi, den, dent, denp
   Real             :: Efn,fn,fnt,fnp,Fnsum
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT
   Real             :: DR,RR,grad
   Real,    Pointer :: Cij(:,:),dCijdtheta(:,:),dCijdphi(:,:)
   Character*100    :: Msg


   Call Start_Timer(Timer(AtomVxc_Timer))
!   Write(msg,*) 'AccumVxc: Error allocating Cij. Size=',Basis_Size
!   Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "AccumVxc:")
   
   Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij
     CoreDen => AT%Core_Density
     CoreGrad => AT%Core_dDendr
     CoreXC => AT%Core_Xchange
     R2 => AT%R2
     DR = AT%Mesh_Step   ! assume that no steps are skipped!
     Ylm => AT%Ylm
     dYlmdtheta => AT%dYlmdtheta
     dYlmdphi => AT%dYlmdphi
     Cij => AT%Cij
     dCijdtheta => AT%dCijdtheta
     dCijdphi => AT%dCijdphi
     FnRad => AT%FnRad
     Mesh_size = AT%Rad_Size


      DR = AT%Mesh_Step*AT%Rad_Skip

     Basis_Size = AT%Basis_Size

     FnSum = 0; ESum = 0

    Do alpha = 1, Num_AngPoints
        Cij = 0
        dCijdtheta=0  ; dCijdphi=0
        Do nili=1, Basis_Size     !*** Calulate Cij for n^a and n~^a
           Li = AT%L_Value(nili)
           Ylm_i => Ylm(:, Li+1, alpha)
           dYlmdt_i => dYlmdtheta(:, Li+1, alpha)
           dYlmdp_i => dYlmdphi(:, Li+1, alpha)
           ibase = nl_Base(nili);  ibase = ibase + nlm_lut(2,ibase);      
           Do njlj=1, Basis_Size
              Lj = AT%L_Value(njlj)
              Ylm_j => Ylm(:, Lj+1, alpha)
              dYlmdt_j => dYlmdtheta(:, Lj+1, alpha)
              dYlmdp_j => dYlmdphi(:, Lj+1, alpha)
              jbase = nl_Base(njlj);  jbase = jbase + nlm_lut(2,jbase);      
           
              den = 0 ; dent = 0 ;   denp = 0
              Do mi = -Li, Li
                 ylm1 = CONJG(Ylm_i(mi+Li+1))
                 ylm1t = CONJG(dYlmdt_i(mi+Li+1))
                 ylm1p = CONJG(dYlmdp_i(mi+Li+1))
              
                 Do mj=-Lj, Lj
                    ylm2 = Ylm_j(mj+Lj+1)
                    ylm2t = dYlmdt_j(mj+Lj+1)
                    ylm2p = dYlmdp_j(mj+Lj+1)
                    den = den + Wij(ibase+mi,jbase+mj) * Ylm1 * Ylm2
                    dent = dent + Wij(ibase+mi,jbase+mj) * &
                         (ylm1t * Ylm2 + ylm1 * ylm2t)
                    denp = denp + Wij(ibase+mi,jbase+mj) * &
                         (ylm1p * Ylm2 + ylm1 * ylm2p)
                 End Do
              End Do

              Cij(nili, njlj) = Den
              dCijdtheta(nili, njlj) = Dent
              dCijdphi(nili, njlj) = Denp
           End Do
        End Do

        !**** Calculate the Radial Densities and store them *****
        FnRad = 0

        Count = 0
        Do nili=1, Basis_Size
           Do njlj=1, Basis_Size
              count = count + 1
              Phi_ij_R2 => AT%Phi_ij_R2(:,count)
              TPhi_ij_R2 => AT%TPhi_ij_R2(:,count)
              dPhijdr => AT%dPhi_ijdr(:,count)
              dTPhijdr => AT%dTPhi_ijdr(:,count)

              FnRad(1:Mesh_Size,1)=FnRad(1:Mesh_Size,1) &          !n
                   + Cij(nili,njlj)*Phi_ij_R2(1:Mesh_Size)

              FnRad(1:Mesh_Size,2)=FnRad(1:Mesh_Size,2) &          !~n 
                   + Cij(nili,njlj)*TPhi_ij_R2(1:Mesh_Size)

              FnRad(1:Mesh_Size,3)=FnRad(1:Mesh_Size,3) &          !dndr
                   + Cij(nili,njlj)*dPhijdr(1:Mesh_Size)

              FnRad(1:Mesh_Size,4)=FnRad(1:Mesh_Size,4) &          !d~ndr 
                   + Cij(nili,njlj)*dTPhijdr(1:Mesh_Size)

              FnRad(1:Mesh_Size,5)=FnRad(1:Mesh_Size,5) &          !dndtheta
                   + dCijdtheta(nili,njlj)*Phi_ij_R2(1:Mesh_Size)

              FnRad(1:Mesh_Size,6)=FnRad(1:Mesh_Size,6) &          !d~ndtheta
                   + dCijdtheta(nili,njlj)*TPhi_ij_R2(1:Mesh_Size)

              FnRad(1:Mesh_Size,7)=FnRad(1:Mesh_Size,7) &          !dndphi
                   + dCijdphi(nili,njlj)*Phi_ij_R2(1:Mesh_Size)

              FnRad(1:Mesh_Size,8)=FnRad(1:Mesh_Size,8) &          !d~ndphi
                   + dCijdphi(nili,njlj)*TPhi_ij_R2(1:Mesh_Size)

           End Do
        End Do

        !*** Add CoreDen  ***
        FnRad(1:Mesh_Size,1) = FnRad(1:Mesh_Size,1) + &
             CoreDen(1:Mesh_Size)
        FnRad(1:Mesh_Size,3) = FnRad(1:Mesh_Size,3) + &
             CoreGrad(1:Mesh_Size)
        FnRad(1:Mesh_Size,2) = FnRad(1:Mesh_Size,2) + &
             AT%CoreTail_Density(1:Mesh_Size)
        FnRad(1:Mesh_Size,4) = FnRad(1:Mesh_Size,4) + &
             AT%GradCoreTail(1:Mesh_Size)

             !**** Calculate Vxc, Gxc,  and Fxc *****
        Do Rk=2, Mesh_Size
           RR=DR*(Rk-1)
           grad=SQRT(FnRad(Rk,3)**2+(FnRad(Rk,5)/RR)**2+(FnRad(Rk,7)/RR)**2)
           Call pbefunc((FnRad(Rk,1)),grad,FnRad(Rk,9),&
                         FnRad(Rk,11),FnRad(Rk,13))
            !!!If (grad>machine_zero) then   ! no longer term has /grad          
            !!!      FnRad(Rk,13)=FnRad(Rk,13)/grad
            !!!Else
            !!!      FnRad(Rk,13)=0.d0
            !!!EndIf
           grad=SQRT(FnRad(Rk,4)**2+(FnRad(Rk,6)/RR)**2+(FnRad(Rk,8)/RR)**2)
           Call pbefunc((FnRad(Rk,2)),grad,FnRad(Rk,10),&
                         FnRad(Rk,12),FnRad(Rk,14))
            !!!If (grad>machine_zero) then
            !!!     FnRad(Rk,14)=FnRad(Rk,14)/grad
            !!!Else
            !!!     FnRad(Rk,14)=0.d0
            !!!EndIf
        End Do

        !FnRad(2:Mesh_Size,15) =   R2(2:Mesh_Size) * &
        !     (FnRad(2:Mesh_Size,9) - CoreXC(2:Mesh_Size) &
        !     - FnRad(2:Mesh_Size,10))
        FnRad(2:Mesh_Size,15) =   R2(2:Mesh_Size) * &
             (FnRad(2:Mesh_Size,9) - FnRad(2:Mesh_Size,10))
        FnRad(1,15)=0.d0

        Efn= IntSimpson(Mesh_Size, DR, FnRad(1:Mesh_Size,15)) * &
              AngPoints(4,alpha)*4*Pi
        Energy = Energy + Efn 

        Count = 0
        Do nili=1, Basis_Size
           Li = AT%L_Value(nili)
           Ylm_i => Ylm(:, Li+1, Alpha)
           dYlmdt_i => dYlmdtheta(:, Li+1, alpha)
           dYlmdp_i => dYlmdphi(:, Li+1, alpha)
           ibase = nl_Base(nili); ibase = ibase + nlm_lut(2,ibase);


           Do njlj=1, Basis_Size
              Lj = AT%L_Value(njlj)
              Ylm_j => Ylm(:, Lj+1, Alpha)
              dYlmdt_j => dYlmdtheta(:, Lj+1, alpha)
              dYlmdp_j => dYlmdphi(:, Lj+1, alpha)
              jbase = nl_Base(njlj); jbase = jbase + nlm_lut(2,jbase);      
              count = count + 1

              Phi_ij_R2 => AT%Phi_ij_R2(:,count)
              TPhi_ij_R2 => AT%TPhi_ij_R2(:,count)
              dPhijdr => AT%dPhi_ijdr(:,count)
              dTPhijdr => AT%dTPhi_ijdr(:,count)

              !**** Radial Integral Portion ******
              FnRad(1,1) = 0; FnRad(Mesh_Size,1) = 0
              FnRad(1,2) = 0; FnRad(Mesh_Size,2) = 0
              FnRad(1,15) = 0; FnRad(Mesh_Size,15) = 0

              Rk=Mesh_Size-1
              
              FnRad(2:Rk,15) =  R2(2:Rk)* &
                   (FnRad(2:Rk,11) * Phi_ij_R2(2:Rk) - &
                    FnRad(2:Rk,12) * TPhi_ij_R2(2:Rk) + &
                    FnRad(2:Rk,13) * FnRad(2:Rk,3) * dPhijdr(2:Rk) - &
                    FnRad(2:Rk,14) * FnRad(2:Rk,4) * dTphijdr(2:Rk))

              FnRad(2:Rk,1) =  &
                   (FnRad(2:Rk,13) * FnRad(2:Rk,5) * Phi_ij_R2(2:Rk) - &
                    FnRad(2:Rk,14) * FnRad(2:Rk,6) * Tphi_ij_R2(2:Rk))

              FnRad(2:Rk,2) =  &
                   (FnRad(2:Rk,13) * FnRad(2:Rk,7) * Phi_ij_R2(2:Rk) - &
                    FnRad(2:Rk,14) * FnRad(2:Rk,8) * Tphi_ij_R2(2:Rk))


              fn = IntSimpson(Mesh_Size, DR, FnRad(1:Mesh_Size,15)) * &
                     AngPoints(4,alpha)*Four_Pi

              fnt= IntSimpson(Mesh_Size, DR, FnRad(1:Mesh_Size,1)) * &
                     AngPoints(4,alpha)*Four_Pi

              fnp = IntSimpson(Mesh_Size, DR, FnRad(1:Mesh_Size,2)) * &
                     AngPoints(4,alpha)*Four_Pi

              Do mi=-Li, Li
                Do mj=-Lj, Lj
                    Dij(ibase+mi,jbase+mj) = Dij(ibase+mi,jbase+mj) + &
                         fn * CONJG(Ylm_i(Li+Mi+1))*Ylm_j(Lj+mj+1) + &
                         fnt*(CONJG(dYlmdt_i(Li+mi+1))*Ylm_j(Lj+mj+1) + &
                             CONJG(Ylm_i(Li+mi+1))*dYlmdt_j(Lj+mj+1)) + &
                         fnp*(CONJG(dYlmdp_i(Li+mi+1))*Ylm_j(Lj+mj+1) + &
                             CONJG(Ylm_i(Li+mi+1))*dYlmdp_j(Lj+mj+1))
                 End Do
              End Do  
           end Do !njlj
        End do !nili
     End Do

  End Do

  Call Stop_Timer(Timer(AtomVxc_Timer))

  Return
End Subroutine
!******************************************************************************
! LoadWij - Temp for comparison
!******************************************************************************

Subroutine LoadWij
   Integer :: i,j,k, nili,njlj,mi,mj,li,lj, ikind, ia, basis_size
   Complex :: c
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:)
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT

   Open(50, file="wij.out", recl=1000)

write(*,*) 'LoadWij:!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'
   Do i=1, Specific_Atoms
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij

     Basis_Size = AT%Basis_Size

     Do j=1, AT%nlm_Size
        Do k=1, AT%nlm_Size
           Read(50,*) ikind,ia,nili,Li,mi,njlj,Lj,mj,c

           Wij(nl_base(nili)+mi+Li,nl_base(njlj)+mj+Lj) = c
        End Do
     End Do

  End Do

  Close(50)
  Return
End Subroutine

!******************************************************************************
! WriteWij - Temp for comparison
!******************************************************************************

Subroutine WriteWij(fd)
   Integer, Intent(IN) :: fd

   Integer :: i,j,k, nili,njlj,mi,mj,li,lj, ikind, ia, Basis_size
   Complex :: c
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:)
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT

!   Open(50, file="wij.out", recl=1000)

   Write(fd,*) 'Printing Wij (nili,Li,mi, njlj,Lj,mj, Wij)********************'
   Do i=1, Specific_Atoms
      Write(fd,*) 'ATOM:',i, '------------------------------------------------'
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij

     Basis_Size = AT%Basis_Size

     Do j=1, AT%nlm_Size
        nili = nlm_LUT(1,j)
        Li = nlm_LUT(2,j)
        mi = nlm_LUT(3,j)

        Do k=1, AT%nlm_Size
           njlj = nlm_LUT(1,k)
           Lj = nlm_LUT(2,k)
           mj = nlm_LUT(3,k)

           c = Wij(nl_base(nili)+mi+Li, nl_base(njlj)+Mj+Lj)
!**           c = c/(4*PI/Xtal%Volume)
           If (ABS(c*COnjg(c))>1E-20) then
              Write(fd,'(6i3)', ADVANCE="NO") nili,Li,mi,njlj,Lj,mj
              Write(fd,*) '(i,j)',nl_base(nili)+mi+Li, nl_base(njlj)+mj+Lj
              Write(fd,*) c
           End If

        End Do
     End Do
  End Do
   Write(fd,*) '******************************************************'
!**     Close(50)
  Return
End Subroutine

!******************************************************************************
! WriteDij - Temp for comparison
!******************************************************************************

Subroutine WriteDij(fd)
   Integer, Intent(IN) :: fd

   Integer :: i,j,k, nili,njlj,mi,mj,li,lj, ikind, ia, Basis_size
   Complex :: c
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:)
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT

!   Open(50, file="wij.out", recl=1000)

  Write(fd,*) 'Printing Dij atom (nili,Li,mi, njlj,Lj,mj, Dij)****************'
   Do i=1, Specific_Atoms
      Write(fd,*) 'ATOM:',i, '------------------------------------------------'
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT
     Wij => A%Wij
     Dij => A%Dij

     Basis_Size = AT%Basis_Size

     Do j=1, AT%nlm_Size
        nili = nlm_LUT(1,j)
        Li = nlm_LUT(2,j)
        mi = nlm_LUT(3,j)

        Do k=1, AT%nlm_Size
           njlj = nlm_LUT(1,k)
           Lj = nlm_LUT(2,k)
           mj = nlm_LUT(3,k)

           c = Dij(nl_base(nili)+mi+Li, nl_base(njlj)+Mj+Lj)
!**           c = c/(4*PI/Xtal%Volume)
           If (ABS(c*COnjg(c))>1E-20) then
              Write(fd,*) 'Dij=',c
              Write(fd,*) '(i,j)',nl_base(nili)+mi+Li, nl_base(njlj)+mj+Lj
              Write(fd,*) '     ',nili,Li,mi,njlj,Lj,mj
           End If

        End Do
     End Do
  End Do
   Write(fd,*) '******************************************************'
!**     Close(50)
  Return
End Subroutine

!******************************************************************************
! WriteQlm - Temp for comparison
!******************************************************************************

Subroutine WriteQlm(Which, fd)
  Integer, Intent(IN) :: Which
   Integer, Intent(IN) :: fd

   Integer :: i,j,k, M,l, ikind, ia, Offset_Q
   Complex :: c, QlmSum, totalQlm
   Complex, Pointer :: Qlm(:,:)
   Integer, pointer :: nl_base(:), nlm_lut(:,:)
   Complex, Pointer :: Dij(:,:), Wij(:,:)
   Type (Specific_Atom), Pointer :: A
   Type (Atom_Info_Fixed), Pointer :: AT

!   Open(50, file="wij.out", recl=1000)


   totalQlm = 0
   If (Which==1) then
      Write(fd,*) 'Printing Qlm (atom, L, M, Qlm)********************'
   Else
      Write(fd,*) 'Printing dEdQlm (atom, L, M, Qlm)********************'
   End If

   Do i=1, Specific_Atoms
      Write(fd,*) 'ATOM:',i, '------------------------------------------------'
     A => Atom_List(i)
     AT => AtomType_Info(A%TypeIndex)
     nl_base => AT%nl_Base
     nlm_LUT => AT%nlm_LUT

     If (Which==1) then
        Qlm => A%Qlm
     else
        Qlm => A%dEdQlm
     End If

     Offset_Q = 2*MaxVal(AT%L_Value) + 1

     QlmSum = 0
     Do L=0, Offset_Q-1
        Do m=-L, L
           c = Qlm(L+1, Offset_Q + m)
           Qlmsum = QlmSum + c
           If (ABS(c*COnjg(c))>1E-5) then
              Write(fd,*) L,M,c
           End If
        End Do
     End Do

     Write(fd,*) ' SUM(Qlm)=',QlmSum
     TotalQlm = TotalQlm + QlmSum
  End Do
  Write(fd,*) '******************************************************'
  Write(fd,*) ' Total SUM(Qlm)=',TotalQlm
  Write(fd,*) '******************************************************'

!**     Close(50)
  Return
End Subroutine


!******************************************************************************
!
!  MakeBandMap - Creates the mapping from diskindex to (Band,Kpnt)
!
!******************************************************************************

Subroutine MakeBandMap
  !Integer :: BandsPerKpnt(NumKpnts), i, j, k
  !Integer :: LUT(Mem_MapSize),  ListSize, SortLUT(Mem_MapSize)
  !Real :: Eigen(Mem_MapSize)
  Integer ::  i, j, k , ListSize
  Integer ,allocatable :: LUT(:), SortLUT(:), BandsPerKpnt(:)
  Real ,allocatable:: Eigen(:)

  
  Allocate(BandsPerKpnt(NumKpnts),LUT(Mem_MapSize),SortLUT(Mem_MapSize), &
            Eigen(Mem_MapSize))
  BandMap = 0
  BandsPerKpnt = 0

  ListSize = 0         !*** Create LUT and Eigenvalue list
  Do i=1, Mem_MapSize
     If (PsiInfo(i)%Kpnt > 0) then
        ListSize = ListSize + 1
        LUT(ListSize) = i
        Eigen(ListSize) = PsiInfo(i)%Energy
     End if
  End Do

           !** Sort the eigenvalues ***
  Call Insertion_Sort(Eigen(1:ListSize), SortLUT(1:ListSize), .TRUE.)  


  Do i=1, ListSize
     j = LUT(SortLUT(i))
     If (PsiInfo(j)%Kpnt > 0) then
        k = PsiInfo(j)%Kpnt
        If (BandsPerKpnt(k) < NumBands) then
           BandsPerKpnt(k) = BandsPerKpnt(k) + 1
           BandMap(1,j) = BandsPerKpnt(k)
           BandMap(2,j) = k
        End If
     End If
  End Do
  DeAllocate(BandsPerKpnt,LUT,SortLUT,Eigen)

  Return
End Subroutine

!******************************************************************************
!
!  CalcNewOccAndFermi - Calculates the New occupancies and Fermi Energy
!            for the system.
!
! ***  NOTE:  This routine assumes the MakeBandMap has been called! ***
!
! Also update virtual memory list
!******************************************************************************

Subroutine CalcNewOccAndFermi
 !Real :: OldOcc(NumBands,NumKpnts), OldEn(NumBAnds,NumKpnts)
  Real, allocatable :: OldOcc(:,:), OldEn(:,:)
  Real :: OldFermi, NewFermi, NewWgt
  Integer :: i, band, Kpnt, temp,ind(1)
 !Integer :: Map(Mem_MapSize), ind(1)
  Integer ,allocatable:: Map(:)

  If (Atomic_Mode == SIM_SEPM) RETURN

  Allocate(OldOcc(NumBands,NumKpnts), OldEn(NumBAnds,NumKpnts), &
             Map(Mem_MapSize))
  OldEn = BZ%KEu       !*** First Map the Energies

  BZ%KEu(:,:) = 10*MaxVal(PsiInfo(:)%Energy)

  Do i=1, Mem_MapSize
     Band = BandMap(1,i)
     Kpnt = BandMap(2,i)

     If ((Kpnt /= 0) .AND. (Band <= NumBands)) then
        BZ%KEu(Band, Kpnt) = PsiInfo(i)%Energy
        PsiInfo(i)%DoSave=1
        PsiInfo(i)%Available=Mem_Used
     Else
        PsiInfo(i)%Kpnt=0
        PsiInfo(i)%DoSave=0
        PsiInfo(i)%Available=Mem_Available
        temp=PsiInfo(i)%MemBufIndex
        if (temp>0) then
           Globalmap%Psi_Handle(temp)%Index=0
           PsiInfo(i)%MemBufIndex=0
        EndIf
     End If
  End Do

  If  (.not.bandstructure_mode) then  !Keep Fermi level if bandstructure calc
    OldFermi = BZ%Fermi
    NewFermi = BZ_CalcFermi(BZ)       !*** Calculate the new fermi energy
    Write(Log_Unit,*) &
         'CalcHewOccAndFermi: NewFermi = ',NewFermi,' OldFermi = ',OldFermi
  EndIf

    If (bandstructure_mode) &
          Write(Log_Unit,*) 'CalcHewOccAndFermi: Keeping OldFermi = ',BZ%Fermi
    
  If  (ReCalculate_Occupancies) then  !Recalculate occupancies                     
    Call BZ_Occupy(BZ, Occupancy)     !*** Occupy the bands
  
    NewWgt = 1.0
    Do i=1, Mem_MapSize      !*** Now Map the new occupations back ***
       If (PsiInfo(i)%Available==Mem_Used) then
          Band = BandMap(1,i)
          Kpnt = BandMap(2,i)
          PsiInfo(i)%Occupancy = (1.0-NewWgt)*PsiInfo(i)%Occupancy + &
                                 NewWgt*Occupancy(Band, Kpnt)
       else
          PsiInfo(i)%Occupancy = 0
       End If
    End Do
  EndIf
  DeAllocate(OldOcc, OldEn,Map)

  Return
End Subroutine


!******************************************************************************
!
!   CalcVxc_LDA - Tabulates the Vxc potential and also accumulates the Exc
!             energy
!
!        Version for Perdew-Wang LDA form only
!
!     Den   - Density in real space
!     Vxc   - Returned tabulated Vxc in FFT space
!
!    NOTE:  Both Den and Vxc use the large grid(G_HIGH)
!
!******************************************************************************

Subroutine CalcVxc_LDA( Den, Vxc)
  Complex, Intent(IN)    :: Den(:)
  Complex, Intent(INOUT) :: Vxc(:)

  Complex, Pointer :: Fn(:)
  Real :: ExcSum, Potxc, Exc, rho, DV
  Integer :: i, N

  ExcSum = 0
  
  !! construct core tail contribution

  Vxc = 0
  Vxc(FFTMap_High(1:Gpnt_Size(G_High)))=SCFvalues%CoreTail(1:Gpnt_Size(G_High))
  Vxc(FFTMap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High))) = &
             Conjg(SCFvalues%CoreTail(2:Gpnt_Size(G_High)))
  Call PerformFFT(FFT_TO_R, G_HIGH, Vxc)

  Do i=1, FFT_Grid(4,G_HIGH)
     rho = Real(Den(i)+Vxc(i))/xtal%volume 
     !!!! -- old form -- Call ExchangeCorr(rho, Potxc, Exc, .FALSE. ,XC_Type)
     Call pwldafunc(rho,Exc,Potxc)
     
     ExcSum = ExcSum + rho*Exc
     Vxc(i) = Potxc 
  End Do

  TildeXC_Energy = TildeXC_Energy + ExcSum*xtal%volume/FFT_Grid(4,G_HIGH)

Write(Log_Unit,*) 'CalcVxc: Vxc_energy=',TildeXC_Energy

  !*** Now convert Vxc to FFT space ***
  write(log_unit,*) 'mag of vxc**2-R',dot_product(vxc,vxc)
  Call PerformFFT(FFT_TO_G, G_HIGH, Vxc)
  write(log_unit,*) 'mag of vxc**2-G',dot_product(vxc,vxc)
  
  SCFValues%VXC(1)=Vxc(FFTMap_High(1))
  SCFValues%VXC(2:Gpnt_Size(G_High))=&
   (Vxc(FFTMap_High(2:Gpnt_Size(G_High)))+&
    Conjg(Vxc(FFTMap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2

  Return
End Subroutine

!******************************************************************************
!
!   CalcVxc_GGA - Tabulates the Vxc potential and also accumulates the Exc
!             energy
!
!        Version for PBE GGA form only
!
!     DG    - Density in G space -- tabulated 1..Gpnt_Size(G_High)
!     W1    - on input: Density in real space, on output: Vxc in FFT space
!     W2    - Work array
!
!    NOTE:  Both W1 and W2 use the large grid(G_HIGH)
!
!******************************************************************************

Subroutine CalcVxc_GGA( W1, W2)   
  Complex, Intent(INOUT) :: W1(:), W2(:)

  Complex, Allocatable :: W3(:),DG(:)
  Real :: ExcSum, Potxc, Exc, rho, DV, grad,fxc,dfxcdn, dfxcdgbg
  Integer :: i, j, N, ier

 
  Allocate(W3(FFT_Grid(4,G_HIGH)),DG(Gpnt_Size(G_HIGH)),stat=ier)
    If (ier /= 0) then
       write(Error_Unit,*) 'CalcVxc_GGA:  Cannot allocate work array ', FFT_Grid(4,G_HIGH)
       stop
    EndIf

  DG = SCFValues%RhoSmooth + SCFValues%CoreTail
  ExcSum = 0

  W2 = 0
  N = Gpnt_Size(G_HIGH)-1
  Do i=1,3
     W3=0
        Do j=2,Gpnt_Size(G_HIGH)
             W3(FFTMap_High(j))=Gpnt(i,j)*DG(j)          
             W3(FFTMap_High(N+j))=-Gpnt(i,j)*CONJG(DG(j))         
        Enddo
     Call PerformFFT(FFT_TO_R, G_HIGH, W3)
     W2 = W2 + W3*CONJG(W3)
  EndDo
           

  Do i=1, FFT_Grid(4,G_HIGH)
     rho = W1(i)/xtal%volume
     grad = SQRT(REAL(W2(i)))/xtal%volume
     call  pbefunc(rho,grad,fxc,dfxcdn,dfxcdgbg)
     
     ExcSum = ExcSum + fxc
     W1(i) = CMPLX(dfxcdn,0)
     W2(i) = CMPLX(dfxcdgbg,0)
  End Do
     W2=W2/xtal%volume

  TildeXC_Energy = TildeXC_Energy + ExcSum*xtal%volume/FFT_Grid(4,G_HIGH)

Write(Log_Unit,*) 'CalcVxc: Vxc_energy=',TildeXC_Energy

  Call PerformFFT(FFT_TO_G, G_HIGH, W1)
    W3=0
    W3(1)=W1(FFTMap_High(1))
        Do j=2,Gpnt_Size(G_HIGH)
           W3(j) = ( W1(FFTMap_High(j)) + CONJG(W1(FFTMap_High(N+j))))/2
        Enddo
  Do i=1,3
     W1=0
        Do j=2,Gpnt_Size(G_HIGH)
             W1(FFTMap_High(j))=Gpnt(i,j)*DG(j)         
             W1(FFTMap_High(N+j))=-Gpnt(i,j)*CONJG(DG(j))        
        Enddo
     Call PerformFFT(FFT_TO_R, G_HIGH, W1)
     W1 = W1*W2
     Call PerformFFT(FFT_TO_G, G_HIGH, W1)
        Do j=2,Gpnt_Size(G_HIGH)
             W3(j)=W3(j)+Gpnt(i,j)*(W1(FFTMap_High(j)) - &
                                    CONJG(W1(FFTMap_High(N+j))))/2
        EndDo
  EndDo
  ! Load resultant W3 into FFT grid as W1
    W1=0
      W1(FFTmap_High(1:Gpnt_Size(G_HIGH))) =  W3(1:Gpnt_Size(G_High))
      W1(FFTmap_High(N+2:Gall_Size(G_HIGH))) =  CONJG(W3(2:Gpnt_Size(G_High)))
           
  write(log_unit,*) 'mag of vxc**2-G',dot_product(W1,W1)

  SCFValues%VXC = W3(1:Gpnt_Size(G_High))
  
  DeAllocate(W3,DG)
  Return
End Subroutine


!******************************************************************************
!
! CalcKineticEnergy - Calculates the KE for the given Psi nad K-pnt
!
!   BlochK - K-Point
!   Psi    - Wave function
!
!******************************************************************************

Real Function CalcKineticEnergy(BlochK, Psi)
  Real,    Intent(IN) :: BlochK(:)
  Complex, Intent(IN) :: Psi(:)

  Integer :: gi, N
  Real    :: G(3), G2
  Complex :: KE


  N = Gpnt_Size(G_LOW) - 1

  KE = 0
  Do gi=1, Gpnt_Size(G_LOW)
     G = BlochK + Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     KE = KE + CONJG(Psi(gi))*Psi(gi) * G2
  End Do
  
  Do gi=2, Gpnt_Size(G_LOW)
     G = BlochK - Gpnt(1:3,gi)
     G2 = DOT_PRODUCT(G,G)
     KE = KE + CONJG(Psi(gi+N))*Psi(gi+N) * G2
  End Do

  CalcKineticEnergy = KE

  Return
End function

!******************************************************************************
!
!  CalcSmoothH - Calculates the smooth density and stores it in V%FAS
!           and also accumulates the energy for the smooth psi to H.
!
!           The KE and Exc for psi~ are ONLY calculated. The terms involving
!           n^ can't be calculated yet since no Wij's and Qlm's are known. 
!
!
!  Both Den and WK1 are LARGE arrays(G_HIGH)
!
!******************************************************************************

Subroutine CalcSmoothH( ProjDotPsi, Den, WK1)
   Complex,  Intent(OUT) :: ProjDotPsi(:,:,:)
   Complex,  Intent(OUT) :: Den(:)
   Complex,  Intent(OUT) :: WK1(:)

   Real     :: G(3), G2, BlochK(3), KE, TotOcc, norm1,norm2
   Type (Mem_handle), Pointer :: Psi
  !Integer          :: Psi_toUse(Mem_MapSize), Kpnt, PsiBand, PsiK,  i
   Integer          ::  Kpnt, PsiBand, PsiK,  i
   Integer  ,allocatable  :: Psi_toUse(:)
   Integer          :: gi, N, j,k, t,Flag
   Complex, Pointer :: Comp(:), Ocomp(:), Work(:), Ox(:)
  !Complex          :: C_PDOT(PLM_MAx)
   Complex  ,allocatable :: C_PDOT(:)
   Logical          :: Stored

   Call Start_Timer(Timer(CalcSmoothH_Timer))

   Allocate(Psi_toUse(Mem_MapSize),C_PDOT(PLM_MAx))
   N = Gpnt_Size(G_LOW) - 1
   
   ProjDotPSi = 0.0

   den = 0

!Write(Log_Unit,*) 'CalcSmoothH:  * Occ=',PsiInfo(:)%Occupancy

!   PsiInfo(:)%PDOT_Stored = .FALSE.  !** Clear PDOT Info

!--- This part taken out by NAWH 6-24-00
!--   !*** Add neutralizing charge if needed ***
!--   If (ABS(NetCharge)>1E-6) then
!--      Stored = .FALSE.
!--      Call GetBuffer( Comp)
!--      Call GetBuffer( OComp)
!--      Comp = 0
!--      Comp(1) = 1
!--      Call OPsi( Comp, Ocomp, C_PDOT, 1, Stored)
!--      ke = DOT_PRODUCT(Comp,Ocomp)
!--Write(Error_Unit, *) 'CalcSmoothH: C_DOT=',ke
!--      Comp = Comp / SQRT(ke)
!--      WK1 = 0
!--
!--      WK1(FFTMap_High(1:Gpnt_Size(G_Low))) = Comp(1:Gpnt_Size(G_Low))
!--      t = Gpnt_Size(G_Low)+1
!--      j = Gpnt_Size(G_HIGH)+1;   k = j + (Gpnt_Size(G_Low)-1) - 1
!--      WK1(FFTMap_High(j:k)) = Comp(t:)
!--
!--WK1 = 0
!--WK1(FFTMap_High(1)) = NetCharge         
!--      Call PerformFFT(FFT_TO_R, G_HIGH, WK1)
!--         
!--      Den = Den + NetCharge*CONJG(WK1)*WK1
!--      
!--      Call FreeBuffer(Comp)
!--      Call FreeBuffer(Ocomp)
!--   End If

   TotOcc = 0
   Do i=1, NumKpnts
      BlochK = BZ%Ku(:,i)

      Psi_toUse = MH_Skip
      Where ((PsiInfo(:)%Kpnt == i) .AND. (PsiInfo(:)%Occupancy >1E-10))
         Psi_toUse = MH_toProcess
      End Where
      Call Phase_Generic( Psi_toUse,  i)
      Call GetNextPsi(Flag, Psi)

      Do While (Flag > 0)
!Write(Log_Unit,*) 'CalcSmoothH: Flag=',Flag, ' * DI=',Psi%Index, ' * Occ=', PsiInfo(Psi%Index)%Occupancy
         TotOcc = TotOcc + PsiInfo(Psi%Index)%Occupancy

         Call FilterValue(Psi%Ptr)

         !*** Calculate the KE for this Psi ***
         KE = CalcKineticEnergy(BlochK, Psi%Ptr)
         PsiInfo(Psi%Index)%KE = KE
         PsiInfo(Psi%Index)%KE_Stored = .TRUE. 

         If (Atomic_Mode == SIM_TIME) then
            Call GetBuffer( Work)
            Call  CalcDEL2_TIME( BlochK, Psi%Ptr, Work)
            KE = DOT_PRODUCT(Psi%Ptr,Work)
            Call FreeBuffer(Work)
         End if

         TildeKE_Energy = TildeKE_Energy + PsiInfo(Psi%Index)%Occupancy*KE


         !*** Now accumulate the density ***
         Call V_G_to_r(Psi%Ptr,G_LOW,WK1,G_HIGH)
         
         Den = Den + PsiInfo(Psi%Index)%Occupancy * CONJG(WK1)*WK1

!**write(Log_Unit,*)'CalcSmoothH',PsiInfo(Psi%Index)%Occupancy
!write(Log_Unit,*) 'den',den(1),den(2),den(3)
!write(Log_Unit,*) 'CalcSmoothH: sum(den) = ',sum(den)*xtal%Volume/PRODUCT(FFT_Grid(1:3,G_HIGH))
         
         !*** Finally calculate the PDOTS for Wij ***
         PsiBand = BandMap(1, Psi%Index)
         PsiK = BandMap(2,Psi%Index)
!Write(Log_Unit,*) 'PsiBand, PsiK=',PsiBand, PsiK, ' * NumBands=',NumBands
         If (.NOT. PsiInfo(Psi%Index)%PDOT_Stored) then 
            If (Proj_Mode == PROJ_RS) then
               WK1 = 0
               Call RS_toR(Psi%Ptr, WK1)
               Call CalcProjProducts(WK1, &
                    PsiInfo(Psi%Index)%PDOT)
            else
               Call CalcProjProducts(Psi%Ptr, PsiInfo(Psi%Index)%PDOT)
            End If

            PsiInfo(Psi%Index)%PDOT_Stored = .TRUE.
         End If


         ProjDotPsi(:,PsiBand, PsiK) = PsiInfo(Psi%Index)%PDOT

!Write(Log_Unit,*) 'CalcSmoothH: PDOT=',ProjDotPsi(:,PsiBand, PsiK)
         
         Call GetNextPsi(Flag, Psi)
      End Do
   End Do

  Write(Log_Unit,*) 'CalcSmoothH: SUM(Occ)=',TotOcc 
  !Den = Den * FFT_Grid(4,G_HIGH)   ! correct for FFT normalization

   DeAllocate(Psi_toUse,C_PDOT)
   Call Stop_Timer(Timer(CalcSmoothH_Timer))
!STOP
   Return
End Subroutine


!******************************************************************************
!
! CalcHam - Calculates the Hamiltonian
!
!******************************************************************************

Subroutine CalcHam( Energy)
   Real,     Intent(OUT) :: Energy

   Integer :: i,j,k, PsiBand, PsiK,  n, t
   Integer          :: den_handle, psi_R_Handle, Psi_G_Handle, KE_Handle
   Complex, Pointer :: BlochK(:), Wk1(:), Wk2(:)
   Real             :: P(3), rmag, Offset(3), energytemp, toosmall
   Real     :: Qt, Qh, c1, c2, gmag
   Complex, Pointer :: OldDen(:), Work1(:), Work2(:), Fn(:), DenTemp(:)
   Complex, Allocatable :: ProjDotPsi(:,:,:)
   Type (Mem_handle), Pointer :: Psi, FAS
  !Integer          :: Psi_toUse(Mem_MapSize), Kpnt, atom, LL, LG, Pin(3)
   Integer          :: Kpnt, atom, LL, LG, Pin(3)
   Integer ,allocatable  :: Psi_toUse(:)
   Complex, Pointer :: Den(:), Vwork(:)
   Real :: Vxc, Exc, ExcSum
   Logical :: Ok, DoZero, DoPrint, DoLoadWij
   Character*100 :: msg
   

   Allocate(Psi_toUse(Mem_MapSize))
   energytemp = 0
   If (Prev_cohesiveEnergy(1) < 1E20) then
      energytemp = Prev_CohesiveEnergy(2) - Prev_CohesiveEnergy(1)
      gmag = 0.5*(abs(Prev_CohesiveEnergy(2)) + abs(Prev_CohesiveEnergy(1)))
      energyTemp = abs(energytemp / gmag)
!      energyTemp = abs(energytemp / Prev_CohesiveEnergy(2))
      If (energytemp > Mix_DampRelative) then
Write(Log_Unit,*) 'DAMP'
         SCFvalues%AC%NewMix = V_DampNewMix
         Do i=1, Specific_Atoms
            Atom_List(i)%Dij_AC%NewMix = Dij_DampNewMix
         End Do
      else
Write(Log_Unit,*) 'RESTORE'
         SCFvalues%AC%NewMix = V_NewMix
         Do i=1, Specific_Atoms
            Atom_list(i)%Dij_AC%NewMix = Dij_NewMix
         End do
      End if
   End If

   Write(Log_unit,*) 'CalcHam: Echange=',energytemp   
   Write(Log_unit,*) 'CalcHam: V_NewMix=',SCFvalues%AC%NewMix
   Write(Log_unit,*) 'CalcHam: Dij_NewMix=',Atom_list(1)%Dij_AC%NewMix, ' *',Dij_NewMix

   Call Start_Timer(Timer(CalcHam_Timer))
   Call PrintDate(Log_Unit, 'CalcHam: Start of Routine!')   
!Energy = 0
!Do i=1, Specific_Atoms
! Energy = Energy +0.5*SUM(Atom_List(i)%Pos**2)
!End Do
!RETURN

DoPrint = .FALSE.
DoLoadWij = .FALSE.
DoZero = .FALSE.


   Diag_Energy = 0
   Hat_Energy = 0
   OneExc_Energy = 0
   Hartree_Energy = 0
   TildeKE_Energy = 0
   TildePot_Energy = 0
   TildeXC_Energy = 0

   Energy = 0

   !*** Allocate the Large FFT and ProjDotPsi arrays ***
   i = FFT_Grid(4,G_HIGH)
   Allocate(Den(i), ProjDotPsi(PLM_MAX, NumBands,NumKpnts), STAT=j)
   msg = 'CalcHam: Could Not allocate LARGE buffers!'
   Call Check_Error(j, msg, Error_Unit, .TRUE.,paw_wc, "CalcHam:")
   !****************

   Vwork =>SCFvalues%Work
If (DoPrint) Write(*,*) 'CalcHam: Start of routine !!!!!!!!!!!!!!!!!!!!!!!!!!'
   Call CalcSmoothH( ProjDotPsi, Den, Vwork)

!write(Log_Unit,*) 'totaldensity after calcsmooth',sum(den)
If (xtal%Rot_Size > 0) Call SymRho( Den, Vwork)
write(Log_Unit,*) 'totaldensity after symrho ',sum(den)

   If (Atomic_Mode /= SIM_AE) then
If (DoPrint) Write(*,*) 'CalcHam: Before Wij!!!!!!!!!!!!!!!!!!!'
      Do i=1, Specific_Atoms
         Atom_List(i)%Wij_Old = Atom_List(i)%Wij
         Atom_List(i)%Dij = 0        !** Init Dij matrix
         Atom_List(i)%Qlm = 0
         Atom_List(i)%dEdQlm = 0
      End Do

      Call CalcWij(ProjDotPsi)    !** Calculate and the Wij's (eq A19)

!Call WriteWij(Log_Unit)
If (DoPrint) then
call writewij(6)
End If
If (DoLoadWij) Call LoadWij

!STOP

If (DoPrint) then
   Write(Log_Unit,*) 'CalcHam: Before SymWij.......................'
   Call WriteWij(Log_Unit)
End If
      If (xtal%Rot_Size > 0)  Call SymWij        !** Symmetrize the Wij's


If (DoPrint) then
   Write(*,*) 'CAlcHam: Before WriteSymWij!!!!!!!!!!!!!!!!!!!!!!!!!!'
   call writewij(6)
   Write(*,*) 'CalcHam: After Wij!!!!!!!!!!!!!!!!!!!'
   !STOP
End If

energytemp = hat_energy
     Call CalcQlm(Hat_Energy)           !** Calculate the Qlm's     (eq A22)
energytemp = hat_energy - energytemp
If (DoPrint) then
  Write(*,*) 'CalcHam: E^ energy =',energytemp
  Write(*,*) 'CalcHam: After CalcQLM'
  Call writeQlm(1,6)
  Call WriteQlm(2,6)
!RETURN
End If

!Call writeQlm(1,Log_Unit)
!Call WriteQlm(2,Log_Unit)

Do i=1, Specific_atoms
  If (DoZero) Atom_List(i)%Dij = CMPLX(0,0)        !** TESTING
End Do
If (DoPrint) Write(*,*) 'CalcHam: Before AddDiag!!!'

      Call AddDiagElements(Diag_Energy)  !** Adds the Diagonal Elements to Dij

If (DoPrint) then 
  Call WriteDij(6)
   
  Do i=1, Specific_atoms
    If (DoZero) Atom_List(i)%Dij = CMPLX(0,0)        !** TESTING
  End Do
End If
       !** Accum Vhat terms in Dij and dEdQlm(eq A23) **
energytemp = hat_energy
     Call AccumVhatdEdQlm(Hat_Energy) 
If (DoPrint) then
   energytemp = hat_energy - energytemp
   Write(*,*) 'CalcHam: V^ energy =',energytemp
   Write(*,*) 'CalcHam: Hat energy =',Hat_energy

   Write(*,*) 'CalcHam: After AccumVhatdEdQlm - dEdQlm'
   Call WriteQlm(2,6)
   Write(*,*) 'CalcHam: After AccumVhatdEdQlm - Dij'
   Call WriteDij(6)

   Do i=1, Specific_atoms
      if (DoZero) Atom_List(i)%Dij = 0        !** TESTING
   End Do
End If
             !** Accumulate Vhartree terms in Dij (eq A26)
      Call AccumVhartree(Hartree_Energy)  
If (DoPrint) then
Write(*,*) 'CalcHam: After AccumVhartree'
Call WriteDij(6)

Write(*,*) 'CalcHam: Before AccumVxc!!!'
Do i=1, Specific_atoms
  If (DoZero) Atom_List(i)%Dij = 0        !** TESTING
End Do
End If
      If (XC_TYPE == XC_LDA_PW) Call AccumVxc_LDA( OneExc_Energy)
      If (XC_TYPE == XC_GGA_PBE) Call AccumVxc_GGA( OneExc_Energy)
If (DoPRint) then
Write(*,*) 'CalcHam: After AccumVxc!!!!!'
Call WriteDij(6)
End If

   Vwork = 0

   If (XC_Type == XC_LDA_PW) then
       Call CalcVxc_LDA( Den, Vwork)   !** Calculate Vxc and Exc 

       !** Convert Density to FFT space ***
       Call PerformFFT(FFT_TO_G, G_HIGH, Den)

       !*** Store the Smooth density *** -- average over n(G)+Conjg(n(-G))
       !SCFvalues%RhoSmooth = Den(FFTmap_High(1:Gpnt_Size(G_High)))
       SCFvalues%RhoSmooth(1) = Den(FFTmap_High(1))
       SCFvalues%RhoSmooth(2:Gpnt_Size(G_High)) = &
            (Den(FFTmap_High(2:Gpnt_Size(G_High))) + &
            Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
    Else if (XC_Type == XC_GGA_PBE) then
       Vwork=Den
       !** Convert Density to FFT space ***
       Call PerformFFT(FFT_TO_G, G_HIGH, Den)

       !*** Store the Smooth density *** -- average over n(G)+Conjg(n(-G))
       SCFvalues%RhoSmooth(1) = Den(FFTmap_High(1))
       SCFvalues%RhoSmooth(2:Gpnt_Size(G_High)) = &
            (Den(FFTmap_High(2:Gpnt_Size(G_High))) + &
            Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
       
       Call CalcVxc_GGA(Vwork,Den)
       Den=0
       Den(FFTMap_High(1:Gpnt_Size(G_High)))=SCFvalues%RhoSmooth(1:Gpnt_Size(G_High))
       Den(FFTMap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High))) = &
             Conjg(SCFvalues%RhoSmooth(2:Gpnt_Size(G_High)))
    Else
       Write(Error_Unit,*) 'CalcHam:    Error in XC_Type ', XC_Type
       stop
    EndIf

!    write(6,*) 'after FFT -- den(1)',den(1)
      Qt = Den(1)

   !** Accumulates the Vlocal term to Vwork                        
   Call AccumVlocal(TildePot_Energy, Den, Vwork)  


     Call CalcDenHat_dEdQlm( Den)     !** Calc N^ 

!!!!!!!!!
!     write(Log_Unit,*)'test \hat{n}', SCFvalues%rhohat(1)

!---------------------
   If (V_MixType == MIX_DENSITY) then
      If (V_NewMix < 0.9999) then 
         Allocate(DenTemp(Gpnt_Size(G_HIGH)))
         !DenTemp = Den(FFTmap_High(1:Gpnt_Size(G_HIGH)))
!  average over n(G) and n(-G)*
          DenTemp(1)=Den(FFTmap_High(1))
          DenTemp(2:Gpnt_Size(G_High)) = &
               (Den(FFTmap_High(2:Gpnt_Size(G_High))) + &
               Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
          DenTemp = DenTemp - &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))

         Call Anderson_Mix(SCFvalues%AC, &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH)), &
              DenTemp)

         Den = 0
         Den(FFTmap_High(1:Gpnt_Size(G_HIGH))) = &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_High))
         Den(FFTmap_High(Gpnt_Size(G_High)+1:)) = &
              CONJG(SCFvalues%Ve_Fourier(2:Gpnt_Size(G_high)))

         DeAllocate(DenTemp)
      else
         SCFvalues%Ve_Fourier(1) = Den(FFTmap_High(1))
         SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)) =  &
              (Den(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
               Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
      End If
   End If


    
if (DoPrint) Call WriteQlm(2,6)
   End IF

   Qh = Den(1) +SCFValues%CoreTail(1) - Qt

   Call AccumShift
   
   Call AccumPotential( TildePot_Energy, Den, Vwork)

!*** New Mix V w/o Vxc and Vlocal
   If (V_MixType == MIX_V) then
      If (V_NewMix < 0.9999) then 
         Allocate(DenTemp(Gpnt_Size(G_HIGH)))
         DenTemp(1) = Den(FFTmap_HIGH(1))
         DenTemp(2:Gpnt_Size(G_HIGH)) = &
             (Den(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
             Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2
         DenTemp(1:Gpnt_Size(G_HIGH)) = DenTemp(1:Gpnt_Size(G_HIGH)) - &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
      
         Call Anderson_Mix(SCFvalues%AC, &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH)), DenTemp)
         
         Den = 0
         Den(FFTmap_HIGH(1:Gpnt_Size(G_HIGH))) = &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
         n = Gpnt_Size(G_High)+Gpnt_size(G_HIGH)-1
         Den(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:n)) = &
              CONJG(SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)))
         Vwork = Vwork + Den
         DeAllocate(DenTemp)
      else
         SCFvalues%Ve_Fourier(1) = Den(FFTmap_High(1))
         SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)) =  &
              (Den(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
               Conjg(Den(FFTmap_High(Gpnt_Size(G_High)+1:Gall_Size(G_High)))))/2

         Vwork(FFTmap_HIGH(1:Gpnt_Size(G_HIGH))) = &
              Vwork(FFTmap_HIGH(1:Gpnt_Size(G_HIGH))) + &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
         n = Gpnt_Size(G_High)+Gpnt_size(G_HIGH)-1
         Vwork(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:n)) = &
              Vwork(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:n)) + &
              CONJG(SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)))
      End If

   End If

   If (Dij_NewMix < 0.9999) then                                        
      Do i=1, Specific_Atoms    !** Mix each atom's Dij separately   
         j = AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size             
         n = (j*(j+1))/2              !**  Determine unique Dij's
                                                                              
         t = 0    !** Copy new Dij's to 1-D mixing array ***         
         Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size
            Do k=1, j                        
               t = t + 1                                     
               Work_Dij(t) = (Atom_List(i)%Dij(j,k)   + &   !** Work_Dij = F(X)
                  CONJG(Atom_List(i)%Dij(k,j)))/2
            End Do
         End Do
                                                                              
         !** Remember that Anderson mixing wants X and F(X)-X ** 
         !** X = Atom_list(i)%Dij_Old and F(X)-X = Work_Dij    
         Work_Dij(1:n) = Work_Dij(1:n) - Atom_list(i)%Dij_old !** Form F(X)-X

         write(Log_Unit,*)'Mix DeltaDij = ',i,Dot_Product(Work_Dij,Work_Dij)
                                                                              
         !*** Perform the mixing ***                             
         Call Anderson_Mix(Atom_List(i)%Dij_AC, Atom_list(i)%Dij_Old, &
              Work_Dij(1:n))                                       
                                                                              
         !*** Now store the results back in the Dij arrays ***      
         !*** Upon exiting Anderson_Mix Atom_List(i)%Dij_old contains the
         !*** mixed or new Dij's to use                              
         t = 0                                                          
         Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size     
            Do k=1, j                                       
               t = t + 1                                      
               Atom_List(i)%Dij(j,k) = Atom_list(i)%Dij_old(t)
               If (j/=k) Atom_List(i)%Dij(k,j) = CONJG(Atom_list(i)%Dij_old(t))
            End Do
         End Do
      End Do
   else
      Do i=1, Specific_Atoms    !** Mix each atom's Dij separately
         t = 0    !** Copy new Dij's to 1-D mixing array ***            
         Do j=1, AtomType_Info(Atom_list(i)%TypeIndex)%nlm_Size
            Do k=1, j                                       
               t = t + 1                          
               Atom_list(i)%Dij_old(t) = (Atom_List(i)%Dij(j,k) + &
                        CONJG(Atom_List(i)%Dij(k,j)))/2
            End Do
         End Do
      EndDo
   End If
                                                                              
!*********  

If (DoPrint) then
   Call WriteWij(Log_Unit)
   Call WriteDij(Log_Unit)
   Call WriteQlm(1,Log_Unit)
   Call WriteQlm(2,Log_Unit)
End If

   If (Atomic_Mode == SIM_AE) then
      Energy = AtomicEnergy - NuclearEnergy - TildePot_Energy &
           - TildeKE_Energy - TildeXC_Energy  

      Write(Log_Unit,*) 'CalcHam: Cohesive Energy = ',Energy
      Write(Log_Unit,*) 'CalcHam:         CalcHam = ',Energy-AtomicEnergy 
      Write(Log_Unit,*) 'CalcHam:   Atomic Energy = ',AtomicEnergy
      Write(Log_Unit,*) 'CalcHam:  Nuclear Energy = ',NuclearEnergy
      Write(Log_Unit,*) 'CalcHam:              KE = ',TildeKE_Energy
      Write(Log_Unit,*) 'CalcHam:      Pot(uncor) = ',TildePot_Energy
      Write(Log_Unit,*) 'CalcHam:        Pot + En = ',TildePot_Energy &
                                                   + NuclearEnergy
      Write(Log_Unit,*) 'CalcHam:              XC = ',TildeXC_Energy
   else
      Energy = TildePot_Energy + TildeKE_Energy + TildeXC_Energy  &
           + OneExc_Energy + Hartree_Energy + Diag_Energy + Hat_Energy

      Write(Log_Unit,*) 'CalcHam: '
      Write(Log_Unit,*) 'CalcHam: Cohesive Energy = ',AtomicEnergy-Energy
      Write(Log_Unit,*) 'CalcHam:         CalcHam = ',Energy 
      Write(Log_Unit,*) 'CalcHam:          Atomic = ',AtomicEnergy 
      Write(Log_Unit,*) 'CalcHam:            Diag = ',Diag_Energy
      Write(Log_Unit,*) 'CalcHam:         Hartree = ',Hartree_Energy
      Write(Log_Unit,*) 'CalcHam:             Hat = ',Hat_Energy
      If (XC_Type == XC_LDA_PW) &
       Write(Log_Unit,*) 'CalcHam:          OneExc = ',OneExc_Energy,' (LDA-PW)'
      If (XC_Type == XC_GGA_PBE) &
       Write(Log_Unit,*) 'CalcHam:          OneExc = ',OneExc_Energy,' (GGA-PBE)'
      Write(Log_Unit,*) 'CalcHam:             KE~ = ',TildeKE_Energy
      Write(Log_Unit,*) 'CalcHam:            Pot~ = ',TildePot_Energy
      If (XC_Type == XC_LDA_PW) &
       Write(Log_Unit,*) 'CalcHam:             XC~ = ',TildeXC_Energy,' (LDA-PW)'
      If (XC_Type == XC_GGA_PBE) &
       Write(Log_Unit,*) 'CalcHam:             XC~ = ',TildeXC_Energy,' (GGA-PBE)'

      Energy = AtomicEnergy - Energy
   End If 

   Write(Log_Unit,*)    'CalcHam:   Smooth Charge = ',Qt

   If (Atomic_Mode /= SIM_AE) then
      Write(Log_Unit,*) 'CalcHam:              Q^ = ',Qh

      rmag = Qt+Qh
      Write(Log_Unit,*) 'CalcHam:           Q~+Q^ =',rmag

      If (ABS(Rmag)>1E-3) then
         Write(Log_Unit,*) 'CalcHam: ERROR!!! Qt+Qh=',rmag
         Write(Error_Unit,*) 'CalcHam: ERROR!!! Qt+Qh=',rmag
      End If
   End If

!   !*** Restrict the potential ***
    If (Filter_potential) then
          Call RestrictPot( Vwork, Den)
       Else
          Den=Vwork
    EndIf

!*** Old V_Mix w/ Vxc and Vlocal ***
   If (V_MixType == MIX_VEFF) then
      If (V_NewMix < 0.9999) then 
         Vwork(1) = Den(FFTmap_HIGH(1))
         Vwork(2:Gpnt_Size(G_HIGH)) = &
           (Den(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
           CONJG(Den(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:Gall_Size(G_HIGH)))))/2
         Den(1:Gpnt_Size(G_HIGH)) = Vwork(1:Gpnt_Size(G_HIGH)) - &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
      
         Write(Log_Unit,*)'Mix DeltaVeff',Dot_Product(Den(1:Gpnt_Size(G_HIGH)),&
             Den(1:Gpnt_Size(G_HIGH)))

         Call Anderson_Mix(SCFvalues%AC, &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH)), &
              Den(1:Gpnt_Size(G_HIGH)))
         
         Den = 0
         Den(FFTmap_HIGH(1:Gpnt_Size(G_HIGH))) = &
              SCFvalues%Ve_Fourier(1:Gpnt_Size(G_HIGH))
         Den(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:)) = &
              CONJG(SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)))
      else
         SCFvalues%Ve_Fourier(1:) = Den(FFTmap_HIGH(1))
         SCFvalues%Ve_Fourier(2:Gpnt_Size(G_HIGH)) =  &
              (Den(FFTmap_HIGH(2:Gpnt_Size(G_HIGH))) + &
              CONJG(Den(FFTmap_HIGH(Gpnt_Size(G_HIGH)+1:Gall_Size(G_HIGH)))))/2
      End If
   End If

   !*** Calc the change in the old and new potentials ***  
   Call PerformFFT(FFT_TO_R, G_HIGH, Den)
!**     write(6,*) 'potential sum' , sum(den)

   Vwork(1:FFT_Grid(4,G_HIGH)) = SCFvalues%Ve
   SCFvalues%Ve = Den(1:FFT_Grid(4,G_HIGH))


   Vwork(1:FFT_Grid(4,G_HIGH))=Vwork(1:FFT_Grid(4,G_HIGH))-SCFvalues%Ve
   gmag = DOT_PRODUCT(Vwork(1:FFT_Grid(4,G_HIGH)),Vwork(1:FFT_Grid(4,G_HIGH)))
   Write(Log_Unit,*) 'CalcHam:     Change in V =',gmag


   Write(Log_Unit,*) ' '
   Write(Log_Unit,*) 'CalcHam:  Change in Wij (atom, change):'
   Qt = 0
   Do i=1, Specific_Atoms
      Atom_List(i)%Wij_Old = Atom_List(i)%Wij - Atom_List(i)%Wij_Old
      gmag = SUM(CONJG(Atom_List(i)%Wij_Old) * Atom_List(i)%Wij_Old)
      Qt = Qt + gmag
      Write(Log_Unit,*) 'CalcHam:CWij      ',i, gmag
   End Do

   Write(Log_Unit,*) 'CalcHam:CWij  Total Change in Wij:', Qt

   DeAllocate(Den,ProjDotPsi)  !*** Free the big arrays ***
    
Call PrintDate(Log_Unit, 'CalcHam: Finished Routine!')   
   Call Stop_Timer(Timer(CalcHam_Timer))

   Prev_CohesiveEnergy(1) = Prev_CohesiveEnergy(2)
   Prev_CohesiveEnergy(2) = energy

   HamLoaded=.true.
   DeAllocate(Psi_toUse)

  Return
End Subroutine


!******************************************************************************
!
!  CalcBandMerit - Calculates the band structure merit function.
!
!******************************************************************************

Subroutine CalcBandMerit( Energy)
   Real,     Intent(OUT) :: Energy

   Integer :: i, Kpnt

   Call Start_Timer(Timer(CalcHam_Timer))

   !!!! Call CalcEigenValues  -- eigenvalues already determined

   Energy = 0
   Do Kpnt=1, NumKpnts
      Do i=1, Mem_MapSize
         If (PsiInfo(i)%Kpnt == Kpnt) then
            If (PsiInfo(i)%Energy <= Eigen_Max) then
               Energy = Energy + ABS(PSiInfo(i)%Energy)
            End If
         End If
      End Do
   end Do

   Call Stop_Timer(Timer(CalcHam_Timer))

   Return
End Subroutine

!******************************************************************************
!
! CalcMeritFunction - Calculates the Merit function
!     Normally CalcHam is called unless BandStructure_Mode = .TRUE. then
!     H is assumed to be fixed and the Band structure merit function is called
! 
!******************************************************************************

Subroutine CalcMeritFunction( Energy)
   Real,     Intent(OUT) :: Energy


   If (BandStructure_Mode) then
      Call CalcBandMerit(Energy)
   else
      Call CalcHam( Energy)
   End If

   Return
End Subroutine

!******************************************************************************
!
!  Hpsi - Multiplies the Hamiltonian times the Given Wave vector
!         and returns the result
!
!  Psi   - Wave vector to multiply. NO Bloch K and DetJ!!!!
!  Hx    - Output result. Hx is initialized to 0!
!  Kpnt  - K-point array index
!
!  KE      - Work Array
!
!******************************************************************************

Subroutine Hpsi( Psi, Hx, ProjDotPsi, Kpnt, PDOT_Stored)
  Complex,          Intent(IN)    :: Psi(:)
  Complex,          Intent(INOUT) :: Hx(:)
  Complex,          Intent(INOUT) :: ProjDotPsi(:)
  Integer,          Intent(IN)    :: Kpnt
  Logical,          Intent(INOUT) :: PDOT_Stored


  Complex, Pointer :: Work(:)
  Integer :: i,j,k

  Call Start_Timer(Timer(HPsi_Timer))

  If (Atomic_Mode == SIM_TIME) then
     Call CalcDEL2_TIME( BZ%Ku(:,Kpnt), Psi, Hx)
  else
     Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), Psi, Hx)
     Hx = -Hx
  End If

  Work => SCFvalues%Work

  If (Atomic_Mode == SIM_SEPM) then
     Call RS_toR(Psi, Work) 
     Work = V_Weight*SCFvalues%Ve*Work
     Call RS_toG(Work, Hx)
  else if (Proj_Mode == PROJ_RS) then
     Call RS_toR(Psi, Work) 
     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Work, ProjDotPsi)
        PDOT_Stored = .TRUE.
     End If

     Work = V_Weight*SCFvalues%Ve*Work
     Call AccumProj(Work, ProjDotPsi, DijMatrix)

     Call RS_toG(Work, Hx)
  else 

     Call V_G_to_r(Psi,G_LOW,Work,G_HIGH)

     Work = V_Weight*SCFvalues%Ve*Work
     Call V_r_to_G(Work,G_HIGH,Hx,G_LOW)

     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Psi, ProjDotPsi)
        PDOT_Stored = .TRUE.
     End If

     Call AccumProj(Hx, ProjDotPsi, DijMatrix)
  End If

  Call Stop_Timer(Timer(HPsi_Timer))

  Return
End Subroutine


!******************************************************************************
!
!  Opsi - Multiplies the Overlap matrix times the Given Wave vector
!         and returns the result
!
!  Psi   - Wave vector to multiply.
!  Ox    - Output result. Ox is initialized to 0!
!  ProjDotPsi - Array containing the projector psi dot products
!  Kpnt - Kpnt array index
!
!******************************************************************************

Subroutine Opsi( Psi, Ox, ProjDotPsi, Kpnt, PDOT_Stored)
  Complex,          Intent(IN)  :: Psi(:)
  Complex,          Intent(OUT) :: Ox(:)
  Complex,          Intent(INOUT) :: ProjDotPsi(:)
  Integer,          Intent(IN)  :: Kpnt
  Logical,          Intent(INOUT) :: PDOT_Stored

  Complex, Pointer :: Work(:)
  COmplex :: xOx
  Real :: temp

!Ox = Psi
!RETURN   !***REMOVE** Debug

  Call Start_Timer(Timer(OPsi_Timer))

  Ox = Psi

  If (Atomic_Mode == SIM_SEPM) then
     !** Don't do anything **
  else If (Proj_Mode == PROJ_RS) then
     Work => SCFvalues%Work
     If (.NOT. PDOT_Stored) then
        Call RS_toR(Psi, Work) 

        Call CalcProjProducts(Work, ProjDotPsi)
        PDOT_Stored = .TRUE.
     End If

     Work = 0
     Call AccumProj(Work, ProjDotPsi, OijMatrix)

     Call RS_toG(Work, Ox)
  else
     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Psi, ProjDotPsi)
        PDOT_Stored = .TRUE.
     End If

     Call AccumProj(Ox, ProjDotPsi, OijMatrix)
  End If

  Call Stop_Timer(Timer(OPsi_Timer))

  Return
End Subroutine

!******************************************************************************
!
!  CalcHxandOx - Calculates H and O times the given Psi
!
!  Psi   - Wave vector to multiply. NO Bloch K and DetJ!!!!
!  Hx    - Output result. Hx is initialized to 0!
!  Ox    - Output result. Ox is initialized to 0!
!  Kpnt  - K-point array index
!
!******************************************************************************

Subroutine CalcHxandOx( Psi, Hx, Ox, PDOT, Kpnt, PDOT_Stored)
  Complex,          Intent(IN)  :: Psi(:)
  Complex,          Intent(OUT) :: Hx(:)
  Complex,          Intent(OUT) :: Ox(:)
  Complex,          Intent(INOUT) :: PDOT(:) 
  Integer,          Intent(IN)  :: Kpnt
  Logical,          Intent(INOUT) :: PDOT_Stored

!*  Complex :: PDOT(PLM_Max)
  Complex, Pointer :: KE(:), Work(:)
  Integer :: i,j


!-----
  Call Start_Timer(Timer(HandOPsi_Timer))

  Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), Psi, Hx)
  Hx = -Hx

  Work => SCFvalues%Work


  If (Atomic_Mode == SIM_SEPM) then
     Call RS_toR(Psi, Work) 
     Work = V_Weight*SCFvalues%Ve*Work
     Call RS_toG(Work, Hx)
     Ox = Psi
  else If (Proj_Mode == PROJ_RS) then
     Call RS_toR(Psi, Work) 
     Ox = Psi
     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Work, PDot)
        PDOT_Stored = .TRUE.
     End If

     RS_Work1 = V_Weight*SCFvalues%Ve*Work
     Work = 0
     Call AccumBothProj(RS_Work1, Work, PDot, DijMatrix, OijMatrix)

     Call RS_toG(RS_Work1, Hx)
     Call RS_toG(Work, Ox)
  else
     Call V_G_to_r(Psi,G_LOW,Work,G_HIGH) 
     Work = V_Weight*SCFvalues%Ve*Work
     Call V_r_to_G(Work,G_HIGH,Hx,G_LOW)
  

     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Psi, PDot)
        PDOT_Stored = .TRUE.
     End If

     Ox = Psi
     Call AccumBothProj(Hx, Ox, PDot, DijMatrix, OijMatrix)
  End If

  Call Stop_Timer(Timer(HandOPsi_Timer))
  Return
End subroutine


!******************************************************************************
!
!  Calc_HOiH_Hx_Ox - Calculates HOiH, H and O times the given Psi
!
!  Psi   - Wave vector to multiply. NO Bloch K and DetJ!!!!
!  HOiHx - Output result. Initialized to 0!
!  Hx    - Output result. Hx is initialized to 0!
!  Ox    - Output result. Ox is initialized to 0!
!  Kpnt  - K-point array index
!
!******************************************************************************

Subroutine Calc_HOiHx_Hx_Ox( Psi, HOiHx, Hx, Ox, PDOT, Kpnt, PDOT_Stored, Eo, Ewidth)
  Complex,          Intent(IN)  :: Psi(:)
  Complex,          Intent(OUT) :: HOiHx(:)
  Complex,          Intent(OUT) :: Hx(:)
  Complex,          Intent(OUT) :: Ox(:)
  Complex,          Intent(INOUT) :: PDOT(:) 
  Integer,          Intent(IN)  :: Kpnt
  Logical,          Intent(INOUT) :: PDOT_Stored
  Real,             Intent(IN)  :: Eo
  Real,             Intent(IN)  :: Ewidth

 !Complex :: PDOT_OiHx(PLM_Max)
  Complex ,allocatable:: PDOT_OiHx(:)
  Complex, Pointer :: KE(:), Work(:), W2(:), W3(:)
  Integer :: i,j


  Allocate(PDOT_OiHx(PLM_Max))
  Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), Psi, Hx)
  Hx = -Hx

  Work => SCFvalues%Work


  If (Proj_Mode == PROJ_RS) then
     Call RS_toR(Psi, Work) 
     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Work, PDot)
        PDOT_Stored = .TRUE.
     End If

     RS_Work1 = V_Weight*SCFvalues%Ve*Work
     Call AccumBothProj(RS_Work1, Work, PDot, DijMatrix, OijMatrix)

     RS_Work2 = Work   !** Copy Ox into a temp space
     Ox = 0
     Call RS_toG(Work, Ox)

     Work = 0
     Call RS_toR(Hx, Work) !** Make sure we add the KE in RS  
     Work = Work + RS_Work1 - Eo*RS_Work2  !** Start the shift

     Call RS_toG(RS_Work1, Hx)
!HOiHx = 0
!Call RS_toG(Work, HOiHx)
!RETURN

     !*** Now calculate HOiHx ***
     Call CalcOinverse(Work, kpnt)   !** Work = OiHx after call
 
     RS_Work1 = Work    !** Need OiHx in both RS and GS

     Call Getbuffer( W2)
     W2 = 0
     Call RS_toG(Work, W2)
     Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), W2, HOiHx)
     HOiHx = -HOiHx
     Call Freebuffer(W2)

     Call CalcProjProducts(RS_Work1, PDot_OiHx)

     RS_Work1 = V_Weight*SCFvalues%Ve*RS_Work1 - Eo*RS_Work1
     Work = 0
     Call AccumBothProj(RS_Work1, Work, PDot_OiHx, DijMatrix, OijMatrix)

     RS_Work1 = RS_Work1 - Eo*Work
     Call RS_toG(RS_Work1, HOiHx)

     HOiHx = HOiHx/Ewidth**2   !** Re-scale the answer     
  else
     Call V_G_to_r(Psi,G_LOW,Work,G_HIGH)
     
     Work = V_Weight*SCFvalues%Ve*Work
     Call V_r_to_G(Work,G_HIGH,Hx,G_LOW)

     If (.NOT. PDOT_Stored) then
        Call CalcProjProducts(Psi, PDot)
        PDOT_Stored = .TRUE.
     End If

     Ox = Psi
     Call AccumBothProj(Hx, Ox, PDot, DijMatrix, OijMatrix)

     !*** Now calculate HOiHx ***
     HOiHx = Hx - Eo*Ox   !** Shift the spectrum
     Call CalcOinverse(HOiHx, kpnt)

!     Call GetBuffer( W3)
!     W3 = HOiHx
     Call CalcProjProducts(HOiHx, PDot_OiHx)

     Call GetBuffer( W2)
     Call CalcDEL2_Phase( BZ%Ku(:,Kpnt), HOiHx, W2)

     Call V_G_to_r(HOiHx,G_LOW,Work,G_HIGH)
     Work = V_Weight*SCFvalues%Ve*Work
     HOiHx = -W2 - Eo*HOiHx
     Call V_r_to_G(Work,G_HIGH,HOiHx,G_LOW)
     Call PerformFFT(FFT_TO_G, G_HIGH, Work)

     W2 = 0
     Call AccumBothProj(HOiHx, W2, PDot_OiHx, DijMatrix, OijMatrix)

     HOiHx = HOiHx - Eo*W2

     HOiHx = HOiHx/Ewidth**2   !** Re-scale the answer

     Call FreeBuffer(W2)
  End If

  DeAllocate(PDOT_OiHx)

  Return
End subroutine

!******************************************************************************
!
!  CalcMagAndEnergy - Calculates the energy and magnitude of the given band
!
!  PsiBand - Wave vector to multiplies band
!  Energy  - Wave function energy
!  Mag     - Magnitude of the wave function
!
!******************************************************************************

Subroutine CalcMagAndEnergy( PsiBand, Energy, Mag)
  Integer,          Intent(IN)    :: PsiBand
  Real,     Intent(OUT)   :: Energy
  Real,     Intent(OUT)   :: Mag

  Complex, Pointer :: Hx(:), Psi(:), KE(:), Ox(:)
 !Complex, TARGET  :: PDot(PLM_Max)
  Complex, TARGET , allocatable  :: PDot(:)
  Integer :: Kpnt
  

  Allocate(PDot(PLM_Max))
  Call GetBuffer( Hx)
  Call GetBuffer( Psi)
  Call GetBuffer( Ox)

  Kpnt = PsiInfo(PsiBand)%Kpnt

  Call GetPsi_fromBuffer( PsiBand, Psi)


  Call CalcHxandOx( Psi, Hx, Ox, &
       PsiInfo(PsiBand)%PDOT, Kpnt, PSiInfo(PsiBand)%PDOT_Stored)

  Energy = Basis_DotProd( Psi,Hx, .FALSE.)
  Mag = Basis_DotProd( Psi,Ox, .FALSE.)

  Call Freebuffer(Psi)
  Call FreeBuffer(Hx)
  Call FreeBuffer(Ox)
  DeAllocate(PDot)

  Return
End Subroutine

!******************************************************************************
!
!  CalcEigenValues - Calculates the Energy eigenvalues for each 
!              wave function 
!
!******************************************************************************
Subroutine CalcEigenValues

 !Integer :: Psi_toUse(Mem_MapSize), i, Flag, Kpnt
  Integer :: i, Flag, Kpnt
  Integer ,allocatable:: Psi_toUse(:)
  Complex, Pointer  :: Hx(:)
  Real              :: DV
  Type(Mem_Handle), Pointer :: Psi
 !Complex           :: PDot(PLM_Max)
  Complex ,allocatable :: PDot(:)


  Allocate(Psi_toUse(Mem_MapSize),PDot(PLM_Max))
  Call Getbuffer( Hx)

  Do Kpnt=1, NumKpnts
     Psi_toUse = MH_Skip
     Where (PsiInfo(:)%Kpnt == Kpnt) Psi_toUse = MH_toProcess
     Call Phase_Generic( Psi_toUse,  Kpnt)

     Call GetNextPsi(Flag, Psi)
     Do While (Flag > 0) 
        i = Psi%Index
        PsiInfo(i)%DoSave=1
        PsiInfo(i)%Available=Mem_Used
        Call Hpsi( Psi%Ptr, Hx, PsiInfo(i)%PDot, Kpnt, &
             PsiInfo(i)%PDOT_Stored)
        PsiInfo(Psi%Index)%Energy = Basis_DotProd(Psi%Ptr, Hx, .FALSE.)
        
!write(Log_Unit,*) 'CalcEigenValues: DI=',Psi%Index, ' * New=',PsiInfo(Psi%Index)%Energy
        Call GetNextPsi(Flag, Psi)
     End Do
  End do

  Call FreeBuffer(Hx)
  DeAllocate(Psi_toUse,PDot)

  Return
End Subroutine

!******************************************************************************
!
!  function spheresymFT
!
!******************************************************************************

Function spheresymFT(ngrid,radialfn,Gmag,dr)
  Real :: spheresymFT
  Integer, INTENT(IN) :: ngrid
  Real, INTENT(IN) :: radialfn(:),Gmag,dr
 
  Real, allocatable :: fn(:)
  Integer       :: i
  Real          :: r, con,x

   con = Four_Pi

   Allocate(fn(ngrid))

   do i=1,ngrid
      r=dr*(i-1)
      x=r
      if (Gmag > machine_precision) x=SIN(Gmag*r)/Gmag
      fn(i) = con*r*radialfn(i)*x
   End do
          
  spheresymFT=IntSimpson(ngrid,dr,fn)

  deallocate(fn)
End Function spheresymFT

Subroutine CalcCoreTail(V)
  Complex, INTENT(OUT) :: V(:)
  Real    :: theta, Gmag, RadG, G(3), Esum,dr
  Complex :: Phase, Vsum
  Integer :: atom, atype, lut, gindex, RadIndex, fi,ngrid,i,j
  Type (Atom_Info_Fixed), Pointer :: AT
 
  Write(Log_unit,*) 'In CalcCoreTail',size(V)
  Do atype=1, Atom_Types
     AT => AtomType_Info(atype)
     Allocate(AT%FTCoreTail(NewMag_Size(G_High)), STAT=j)
        if (j /= 0) then
           write(Error_unit,*) 'Error CalcCoreTail: allocation ', &
                    NewMag_Size(G_High),atype,j
           stop
        endif
     AT%FTCoreTail=0
     dr= AT%Mesh_Step
     ngrid=AT%CoreTail_Points
     Radindex=0
     Do gindex=1, Gpnt_Size(G_HIGH)
         If (G_NewMag(gindex)) then
            Gmag = Gpnt(4,gindex)
            RadG = spheresymFT(ngrid,AT%CoreTail_Density,Gmag,dr)
            RadIndex=RadIndex+1
            AT%FTCoreTail(RadIndex)=RadG
         Endif
     Enddo
  Enddo
    
  V=0; Radindex=0
  Do gindex=1, Gpnt_Size(G_HIGH)
     G =  Gpnt(1:3,gindex)
     If (G_NewMag(gindex)) Radindex=Radindex+1
 
     Vsum = 0
     Do atype=1, Atom_Types
        AT => AtomType_Info(atype)
                                                                                
        !*** Create the structure factors and accumulate vlocal***
        Do lut = AtomType_Range(1,atype), AtomType_range(2,atype)
           atom = Atomtype_map(lut)
                                                                                
           theta = DOT_PRODUCT(G, Atom_List(atom)%Pos)
           Phase = CMPLX(cos(theta), -sin(theta))
           Vsum = Vsum + Phase*AT%FTCoreTail(RadIndex)
        End Do
     End do
      
     V(gindex)=Vsum
  End Do
End Subroutine


!******************************************************************************
!
!  CalcRadial_Vlocal - Calculates the radial Vlocal grids
!
!******************************************************************************

Subroutine CalcRadial_Vlocal
 
  Integer       :: i,j, k, G, Gmag, n, L,ngrid
  Character*200 :: Token
  Real, Pointer :: RadV(:),fn(:)
  Real          :: dr, r,  norm,xmax,gmax,con
  Type (Atom_Info_Fixed), Pointer :: AT

   con = Four_Pi/Xtal%volume
   L=0  ! L=0 for Vlocal

   ngrid=MAXVAL(AtomType_Info(:)%Mesh_Size)
   Allocate(fn(ngrid))

  Do i=1, Atom_Types
    AT => AtomType_Info(i)

    Allocate(AT%Rad_Vlocal(NewMag_Size(G_High)), STAT=j)
    Write(Token,*) 'CalcRadial_Vlocal: Error allocating array. Size=',&
         NewMag_Size(G_HIGH), ' * TypeIndex=',i
    Call Check_Error(j, Token, Error_Unit, .TRUE., PAW_WC,'CalcRadial_Vlocal:')

    RadV => AT%Rad_Vlocal

    RadV = 0

    If (Atomic_Mode == SIM_SEPM) then 
       k = 0
       con = 1.0 /xtal%volume
       Do j=1, Gpnt_Size(G_HIGH)
          If (G_NewMag(j)) then
             k = k + 1
             RadV(k) = EvalGauss(AT%V_sepm, Gpnt(4,j)) * con
          End If
       End Do
    else
      ! norm = con*AT%V_Local
       norm = con
       
       if (ABS(norm).gt.1.d-9) then
          
          DR=AT%Mesh_Step      
          write(6,*) 'CalcRadial_Vlocal: DR',DR
          
          !*** Tabulate the vlocal function ***
          Do j=1, AT%Mesh_Size
             r = DR*(j-1)
             
             !Fn(j) = norm*r*AT%Shape_Func(j)
             Fn(j) = norm*r*AT%RadR_Vlocal(j)
          End do
          
          k = 0
          Do j=1, Gpnt_Size(G_HIGH)
             If (G_NewMag(j)) then
                k = k + 1
                !RadV(k) = RadialFourier(L, Gpnt(4,j), DR, AT%Mesh_Size, fn)
                RadV(k) = spheresymFT(AT%Mesh_Size,&
                    AT%RadR_Vlocal,Gpnt(4,j),dr)/xtal%volume
             End If
          End Do
       Endif
    End If
  End Do

  Deallocate(fn)

  Return
End Subroutine


!******************************************************************************
!
!  CalcRadialHat - Calculates the radial n^ grids
!
!******************************************************************************

Subroutine CalcRadialHat

  Integer       :: i,j, k, G, Gmag, n, L, ngrid
  Character*200 :: Token
  Real, Pointer :: RadHat(:),fn(:)
  Real          ::  dr, r,  norm,xmax,gmax
  Type (Atom_Info_Fixed), Pointer :: AT


   ngrid=MAXVAL(AtomType_Info(:)%Mesh_Size)
   Allocate(fn(ngrid))


  Do i=1, Atom_Types
    AT => AtomType_Info(i)

    AT%Hat_MaxL = 2*MaxVal(AT%L_Value)
  
    Allocate(AT%RadHat(NewMag_Size(G_High), AT%Hat_MaxL+1), STAT=j)
    Write(Token,*) 'CalcRadialHat: Error allocating array. Size=',&
         NewMag_Size(G_HIGH), ' * ', AT%Basis_Size, ' * TypeIndex=',i
    Call Check_Error(j, Token, Error_Unit, .TRUE., PAW_WC, 'CalcRadialHat:')
    
    DR=AT%Mesh_Step    
     !write(6,*) 'CalcRadHat: DR',DR

    Do L=0, AT%Hat_MaxL
       RadHat => AT%RadHat(:,L+1)

       n=AT%Mesh_Size
       do j=1,n            
         fn(j)=((dr*(j-1))**(2*L+2))*AT%Shape_Func(j)
       enddo

       norm = sqrt(Four_Pi)/IntSimpson(n, DR, fn)

       !*** Tabulate the hat function ***
       Do j=1, n
          r = DR*(j-1)
          Fn(j) = norm * AT%Shape_Func(j) * r**(L+1)
       End do

       k = 0
       Do j=1, Gpnt_Size(G_HIGH)
          If (G_NewMag(j)) then
             k = k + 1
             RadHat(k) = RadialFourier(L, Gpnt(4,j), DR, n, fn)
            !write(6,'("radhat",2i4,1p2e15.7)') k,L,Gpnt(4,j),radHat(k)
          End If
       End Do
    End Do

  End Do

  Deallocate(fn)

  Return
End Subroutine


!******************************************************************************
!
!  PrepareVxc - Prepares the data structures for the caclulation of Vxc
!
!  AT - Pointer to the Atom_Info_Fixed structure
!
!******************************************************************************

Subroutine PrepareVxc(AT)
  Type (Atom_Info_Fixed), Intent(INOUT) :: AT
  
  Real, Pointer :: R2(:), Phi_ij_R2(:,:), TPhi_ij_R2(:,:)
  Real, Pointer :: dPhi_ijdr(:,:),dTPhi_ijdr(:,:)
  Real, Pointer :: Phi_I(:), Phi_j(:), TPhi_i(:), TPhi_j(:)
  Complex, Pointer :: Ylm(:,:,:),dYlmdtheta(:,:,:),dYlmdphi(:,:,:)
  Complex :: Ylm_i(13), dYlmdt(13),dYlmdp(13)
  Real      :: DR,  Ra(3),den,Exc,Vxc,grad,dfxcdn,dfxcdg
  Integer   :: RadSize, Rk, alpha, count, nili, njlj, mi, mj, Li, Lj, Stride


  RadSize = AT%Rad_Size

  !******* First determine how many ij indices there are *******
  count = 0
  Do nili=1, AT%Basis_Size 
     Li = AT%L_Value(nili)
     Do njlj=1, AT%Basis_Size
        Lj = AT%L_Value(njlj)

        count = count + 1
     End Do
  End Do


  Stride = AT%Rad_Skip

  Li = MAXVal(AT%L_Value)
  mi = 2*Li+1
  Allocate(AT%R2(RadSize), &
       AT%Ylm(mi, Li+1, Num_AngPoints), AT%Phi_ij_R2(RadSize, Count), &
       AT%TPhi_ij_R2(RadSize, Count), AT%Cij(AT%Basis_Size, AT%Basis_Size), &
       STAT=Rk)
  If (XC_Type == XC_LDA_PW) Allocate(AT%FnRad(RadSize,7))
  If (XC_Type == XC_GGA_PBE) then
    Allocate(AT%FnRad(RadSize,15),AT%dYlmdtheta(mi, Li+1, Num_AngPoints),&
         AT%dYlmdphi(mi, Li+1, Num_AngPoints), &
         AT%dCijdtheta(AT%Basis_Size, AT%Basis_Size),&
         AT%dCijdphi(AT%Basis_Size, AT%Basis_Size),&
         AT%dPhi_ijdr(RadSize, Count),&
         AT%dTPhi_ijdr(RadSize, Count),&
         AT%Core_dDendr(RadSize),AT%GradCoreTail(RadSize),STAT=Rk)
  EndIf
 !**** fix core part -- this was previously in InitSystem
     !**** Create the Core XC ********
        If (XC_Type == XC_LDA_PW) then
          ! Do Rk=2, AT%Mesh_Size
          !    den = AT%Core_Density(Rk)
          !    If (XC_Type == XC_LDA_PW) Call pwldafunc(den,Exc,Vxc)
          !    AT%Core_Xchange(Rk) = Exc
          ! End Do
          ! !AT%Core_Xchange(1) = 2*AT%Core_Xchange(2)-AT%Core_Xchange(3)
          ! AT%Core_Xchange(1) = 3*(AT%Core_Xchange(2)-AT%Core_Xchange(3))&
          !             +  AT%Core_Xchange(4)
        Else If (XC_Type == XC_GGA_PBE) then
           RadSize=AT%Mesh_Size
           Call nderiv(AT%Mesh_Step,AT%Core_Density(:),&
                 AT%Core_dDendr(:),RadSize)
           Call nderiv(AT%Mesh_Step,AT%CoreTail_Density(:),&
                 AT%GradCoreTail(:),RadSize)
          ! Do Rk=1,RadSize
          !  grad=ABS(AT%Core_dDendr(Rk))     ! stores gradient
          !  Call  pbefunc(AT%Core_Density(Rk),grad,AT%Core_Xchange(Rk),dfxcdn,dfxcdg)
          !            !! Remember: for GGA Core_Xchange is fxc
          ! Enddo
        Else
          Write(Error_Unit, * ) 'Initsystem:  Error in XC_Type', XC_Type
          Stop
        EndIf
 !**** end core part
  R2 => AT%R2
  Ylm => AT%Ylm
   If (XC_Type == XC_GGA_PBE) dYlmdtheta => AT%dYlmdtheta
   If (XC_Type == XC_GGA_PBE) dYlmdphi => AT%dYlmdphi
   If (XC_Type == XC_GGA_PBE) dPhi_ijdr => AT%dPhi_ijdr
   If (XC_Type == XC_GGA_PBE) dTPhi_ijdr => AT%dTPhi_ijdr

  !********** Make R^2 ************
  DR = AT%Mesh_Step*Stride !** Calc Radial weighting and R
  Do Rk=1, RadSize
     R2(Rk) = (DR*(Rk-1))**2
  End Do

  count = 0
  Do nili=1, AT%Basis_Size
     Li = AT%L_Value(nili)
     Phi_i =>AT%Phi(1:AT%Mesh_Size:Stride,nili)
     TPhi_i =>AT%TPhi(1:AT%Mesh_Size:Stride,nili)

     Do Rk=1, Num_AngPoints  !** Calculate the Ylm's **
        Ra = AngPoints(1:3, Rk)
        Ylm_i = Spharm(Ra(1), Ra(2), Ra(3), Li, .TRUE.)
        Ylm(1:2*Li+1, Li+1, Rk) =  Ylm_i(1:2*Li+1)
      If (XC_Type == XC_GGA_PBE) then
        Call GradYlm(Ra(1),Ra(2),Ra(3),Li,dYlmdt,dYlmdp)
        dYlmdtheta(1:2*Li+1, Li+1, Rk) =  dYlmdt(1:2*Li+1)
        dYlmdphi(1:2*Li+1, Li+1, Rk) =  dYlmdp(1:2*Li+1)
      EndIf
     End Do

     Do njlj=1, AT%Basis_Size
        count = count + 1

        Phi_j =>AT%Phi(1:AT%Mesh_Size:Stride,njlj)
        TPhi_j =>AT%TPhi(1:AT%Mesh_Size:Stride,njlj)

        Phi_ij_R2 => AT%Phi_ij_R2
        TPhi_ij_R2 => AT%TPhi_ij_R2

        Phi_ij_R2(2:RadSize,count) = &
                 Phi_i(2:RadSize)*Phi_j(2:RadSize)/R2(2:RadSize)
        !Phi_ij_R2(1,count)=2*Phi_ij_R2(2,count)-Phi_ij_R2(3,count)
        Phi_ij_R2(1,count)=3*(Phi_ij_R2(2,count)-Phi_ij_R2(3,count)) &
                             + Phi_ij_R2(4,count)
        TPhi_ij_R2(2:RadSize,count) = &
                 TPhi_i(2:RadSize)*TPhi_j(2:RadSize)/R2(2:RadSize)
        !TPhi_ij_R2(1,count)=2*TPhi_ij_R2(2,count)-TPhi_ij_R2(3,count)
        TPhi_ij_R2(1,count)=3*(TPhi_ij_R2(2,count)-TPhi_ij_R2(3,count)) &
                             + TPhi_ij_R2(4,count)
       If (XC_Type == XC_GGA_PBE) then
        call nderiv(DR,Phi_ij_R2(:,count),dPhi_ijdr(:,count),RadSize)
        call nderiv(DR,TPhi_ij_R2(:,count),dTPhi_ijdr(:,count),RadSize)
       EndIf

     End Do
  End Do

  Return
End Subroutine

!******************************************************************************
!
! InitHam - Initializes the HAmiltonian module
!
!******************************************************************************

Subroutine InitHam
  Integer :: i,j, Rad_Size
  Type (atom_info_fixed), Pointer :: AT
  Complex, Pointer :: Work(:), Work2(:)

  Prev_CohesiveEnergy = 1E30

  Allocate(BandMap(2,Mem_MapSize),stat=i)
    If (i /= 0) then
      Write(Error_Unit,*) 'InitHam:  Failed to allocate BandMap',i
      Call Flush(Error_Unit)
      stop
    EndIf

  If (Atomic_Mode == SIM_AE) RETURN   !** Ignore the rest if an AE calc

!RETURN !**REMOVE** Debug Only

       !** Calculate the radial hat functions
  If (Atomic_Mode /= SIM_SEPM) Call CalcRadialHat      
  Call CalcRadial_Vlocal  !** Calculate the radial V_local functions
  Call CalcVlocal(SCFvalues%V_local)
  Call CalcCoreTail(SCFvalues%CoreTail)

  write(Log_unit,*) 'InitHam: SCFvalues%CoreTail', SCFvalues%CoreTail(1)
  
  If (Atomic_Mode == SIM_SEPM) then
     Call GetBuffer( Work2)

     Work => SCFvalues%V_Local
     Work2(1:Gpnt_Size(G_LOW)) = Work(1:Gpnt_Size(G_Low))
     SCFvalues%Ve_Fourier =  Work(1:Gpnt_Size(G_LOW))

     Work2(Gpnt_Size(G_LOW)+1:) = &
          Work(Gpnt_Size(G_high)+1:Gpnt_Size(G_High)+Gpnt_size(G_low)-1)     

     Work => SCFvalues%Work
     Work = 0
     Call RS_toR(Work2, Work)
     SCFvalues%Ve = Work

!SCFvalues%Ve = 0
     Call FreeBuffer(Work2)

     RETURN  !*** EXIT EARLY ***
  End IF

  Do i=1, Atom_Types
     AT => AtomType_Info(i)
     Allocate(AT%Rad_Size, AT%Rad_Skip)

!AT%Mesh_Size = At%Mesh_Size - 2   !*** DEBUG ONLY ****
!AT%Mesh_Size = At%Mesh_Size-1  !*** DEBUG ONLY * for li


     Rad_Size = 0
     Do j=1, AT%Mesh_Size, AtomicVxc_Step
        Rad_Size = Rad_Size + 1
     End Do
     AT%Rad_Size = Rad_Size
     AT%Rad_Skip = AtomicVxc_Step     !note this is always 1 unless
                                      !  we fix the Core_Density part!!!
     Write(Log_Unit,*) 'InitHam: ',' * Atom=',i, &
          ' * SIZE=',AT%Rad_Size, &
          ' * SKIP=',AT%Rad_Skip     

  End Do


  !**** Prepare the Vxc data structures *****
     !************ Generate the Angular Integration points ***************
     !Call InitSpharm
     Num_AngPoints = Angular_Points**2
     AngPoints => AngInit(Angular_Points,Angular_Points,Xtal%Volume,Error_Unit)
  Do i=1, Atom_Types
     AT => AtomType_Info(i)
     Call PrepareVxc(AT)
  End Do

  !*** Allocate space for Dij mixing ***
  i = MaxVal(AtomType_Info(:)%nlm_size)
  i = i*(i+1)/2
  Allocate(Work_Dij(i),stat=j)
    If (j /= 0) then
      Write(Error_Unit,*) 'InitHam:  Failed to allocate BandMap',j
      Call Flush(Error_Unit)
      stop
    EndIf


  !*** Init the Dij Mixing ***
  Do i=1, Specific_Atoms
     j = AtomType_Info(Atom_List(i)%TypeIndex)%nlm_Size
     j = j*(j+1)/2
     Call InitAnderson(Atom_List(i)%Dij_AC, Error_Unit, Mix_Size, &
          j, 1E-5, Dij_NewMix, AndersonConditionNo)
  End Do

  Return
End Subroutine

End Module
