!******************************************************************************
!
! File : timedependent.f90
!   by : Alan Tackett
!   on : 10/22/99
!  for : PWPAW
!
!  Contains routines to propagate the wave functions for a time-dependent
!  calculation.
!
!******************************************************************************

Module timedependent

Use paw_inout
Use misc
Use options_data
Use word
Use atom_data
Use psilib
Use solver
Use mem_data
Use memmgr
Use work_mgr
Use storedata

Implicit NONE!!!!!!!!!

Real, PARAMETER :: au_a0 = 0.5291772083E0
Real, PARAMETER :: au_eV = 13.60569174E0
Real, PARAMETER :: au_t0 = 0.02418884326500E0


!******************************************************************************
Contains
!******************************************************************************

!******************************************************************************
!
! LdElectricField - Loads the electric field from the input
!
!     Input format:
!         
!        ELECTRIC_FIELD
!            TYPE  SINE
!            E0  Ex Ey Ez   # Volts / bohr
!            T0  start_time # Femto-seconds
!            PHASE phase    # radians
!            FREQUENCY  f   # Hz
!        END
!
!******************************************************************************

Subroutine LdElectricField(WC)
  Type (Word_Context), Intent(INOUT) :: WC

  Integer       :: i,j, tlen
  Character*100 :: token, token2

  E_Field%EType = -1
  E_field%E0 = 1E30
  E_Field%T0 = 0
  E_Field%Phase = 0
  E_Field%Frequency = 1E30
  
  Call GetNextWord(WC, token, Tlen)
  Call UpperCase(Token)

  Do While (token /= 'END')
    If (TRIM(token) == 'TYPE') then
       Call GetNextWord(WC, token, tlen)
       Call UpperCase(token)
       If (TRIM(token) == 'SINE') then
          E_field%Etype = E_SINE
       else
          Write(Error_Unit,*) 'LdElectricfield: Unknown Type! Read :', &
               TRIM(token)
          Write(Error_Unit,*) 'LdElectricField: Should be SINE.'
          STOP
       End IF
    else if (TRIM(token) == 'E0') then
       Call GetNumbers(WC, E_field%E0)
    else If (TRIM(token) == 'T0') then
       Call GetNumber(WC, E_field%T0)
    else if (TRIM(token) == 'PHASE') then
       Call GetNumber(WC, E_field%phase)
    else if (TRIM(token) == 'FREQUENCY') then
       Call GetNumber(WC, E_field%Frequency)
    else
       Write(Error_Unit,*) 'LdElectricfield: Unknown option! Read:',TRIM(token)
       Write(Error_Unit,*) 'LdElectricField: Ignoring option.'
    End If

    !** Get next command **
    Call GetNextWord(WC, token, Tlen)
    Call UpperCase(Token)
  End Do

  Return
End Subroutine
 
!******************************************************************************
!
!  CalcVectorPotential - Calculates the vector potential and returns the
!             vector at the given time.
!
!      T - Current time in Femto-sectonds
!
!******************************************************************************

Subroutine CalcVectorPotential(t, Elec, Vec_Pot)
  Real, Intent(IN)  :: t
  Real, Intent(OUT) :: Elec(:)
  Real, Intent(OUT) :: Vec_Pot(:)

  Real :: phase, w

  w = 2*PI*E_field%Frequency*au_t0*1E-15
  
  phase = w*(t + E_field%t0)/au_t0 + E_field%Phase

  Vec_Pot = E_field%E0*(au_a0/au_ev) / w * cos(phase)

  Elec = E_Field%E0*sin(phase)*(au_a0/au_ev)

  Return
End Subroutine

!******************************************************************************
!
!  TIME_PropagatePsi - Propagates the given wave function to the next
!            time step.
!
!     Psi     - Wave function to update
!     NewPsi  - New Wave function
!     Kpnt    - Kpnt for Psi
!
!******************************************************************************

Subroutine TIME_PropagatePsi(PsiIndex, Psi, NewPsi, Kpnt)
  Integer, Intent(IN)  :: PsiIndex
  Complex, Intent(IN)  :: Psi(:)
  Complex, Intent(OUT) :: NewPsi(:)
  Integer, Intent(IN)  :: Kpnt

  Complex, PARAMETER :: ani = (0,-1)
  integer :: i, j
  Complex, Pointer :: Curr(:), Prev(:), Temp(:), Base1(:), Base2(:)
  Complex :: PDOT(PLM_Max), coeff, phase
  Logical :: PDOT_Stored
  Real :: dot, dtn


  Call GetBuffer( Base1)
  Call GetBuffer( Base2)

  Curr => Base1;  Prev => Base2

  PDOT_Stored = PsiInfo(PsiIndex)%PDOT_Stored
  PDOT = PsiInfo(PsiIndex)%PDOT

Curr = 0
phase = 1
  i = 0;
  coeff = 1
  dtn = 1
  Prev = Psi
  NewPsi = Psi
  dot = 2*TIME_TOL
!Write(Log_Unit,*) 'TIME_Prop:  dot=',dot, ' * TIME_TOL=',TIME_TOL, ' * Time_MaxN=',Time_MaxN

!return
  Do while ((dot>TIME_TOL) .AND. (i<TIME_MaxN))
     i = i + 1
!Curr = Psi * PsiInfo(PsiIndex)%Energy**i

     Call HPsi( Prev, Curr, PDOT, Kpnt, PDOT_Stored)
     PDOT_Stored = .FALSE.
     Call CalcOinverse(Curr, kpnt)

     coeff = coeff * ani / i
     dtn = dtn * TimeStep/au_t0
     NewPsi = NewPsi + Curr*coeff*dtn
phase = phase + coeff*dtn*(PsiInfo(PsiIndex)%Energy**i)

     dot = DOT_PRODUCT(Curr,Curr) *CONJG(coeff)*coeff*dtn**2
     dot = SQRT(ABS(dot))
     !*** Swap Curr and Prev ***
!Write(Log_Unit,*) 'TIME_Prop: DI=',PsiIndex, ' * i=',i, ' * prev=',DOT_PRODUCT(Prev,Prev),' * curr=',DOT_PRODUCT(Curr,Curr), ' * dot=',dot
     temp => Curr
     Curr => Prev
     Prev => temp
  End Do

  Write(Log_Unit,*) 'TIME_Prop: DI=',PsiIndex, ' * i=',i, ' * dot=',dot

dot = -PsiInfo(PsiIndex)%Energy*TimeStep/au_t0
coeff = CMPLX(cos(dot), sin(dot))
Write(Log_Unit,*) 'TIME_Prop:             CalcP=',phase
Write(Log_Unit,*) 'TIME_Prop:             phase=',coeff, ' * E=',PsiInfo(PsiIndex)%Energy
PDOT_Stored = PsiInfo(PsiIndex)%PDOT_Stored
PDOT = PsiInfo(PsiIndex)%PDOT
Call Opsi( Psi, Curr, PDOT, Kpnt, PDOT_Stored)
!Curr = Psi - NewPsi
coeff = DOT_PRODUCT(Curr,NewPsi)
Write(Log_Unit,*) 'TIME_Prop:    <Psi|O|NewPsi>=',coeff

PDOT_Stored = .FALSE.
Call Opsi( NewPsi, Curr, PDOT, Kpnt, PDOT_Stored)
coeff = DOT_PRODUCT(NewPsi,Curr)
Write(Log_Unit,*) 'TIME_Prop:    <NewPsi|O|NewPsi>=',coeff

PDOT_Stored = .FALSE.
Call HPsi( NewPsi, Curr, PDOT, Kpnt, PDOT_Stored)
coeff = DOT_PRODUCT(NewPsi,Curr)
Write(Log_Unit,*) 'TIME_Prop:    <NewPsi|H|NewPsi>=',coeff

  Call FreeBuffer(Base1)
  Call FreeBuffer(Base2)


  Return
End Subroutine


!******************************************************************************
!
! CalcTotalCurrent - Calculates the total current in the supercell
!
!******************************************************************************

Subroutine CalcTotalCurrent(Current)
  Real, Intent(OUT)  :: Current(3)

  Type (Mem_Handle), Pointer :: Psi
  Integer :: Psi_toUse(Mem_MapSize),  flag, i, N, gi,kpnt
  Real :: G(3)
  Complex :: CurrSum(3)

  Psi_toUse = MH_Skip
  Where (PsiInfo(:)%Occupancy > 1E-10) Psi_toUse = MH_toProcess

  kpnt = 1  !** Not actually used since don't have to do any PDOT's
  Call Phase_Generic( Psi_toUse,  kpnt)

  N = Gpnt_Size(G_LOW)-1
 
  CurrSum = 0
  Call GetNextPsi(Flag, Psi)
  Do While (Flag /= 0)
     Kpnt=PsiInfo(Psi%Index)%Kpnt
     Do gi=1, Gpnt_Size(G_LOW)
        G = Gpnt(1:3,gi) + BZ%Ku(1:3,Kpnt) - Vector_Pot
        CurrSum = CurrSum + G*CONJG(Psi%Ptr(gi))*Psi%Ptr(gi)
     End Do
     Do gi=2, Gpnt_Size(G_LOW)
        G = - Gpnt(1:3,gi) + BZ%Ku(1:3,Kpnt) - Vector_Pot
        CurrSum = CurrSum + G*CONJG(Psi%Ptr(gi+N))*Psi%Ptr(gi+N)
     End Do
     
     Call GetNextPsi(Flag, Psi)
  End Do

  Current = CurrSum

  Return
End Subroutine



End Module
